# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx
import dwave_networkx as dnx
from dimod import ExactCQMSolver


class TestPartitioning(unittest.TestCase):
    def test_edge_cases(self):
        # get the empty graph
        G = nx.Graph()

        node_partitions = dnx.partition(G, sampler=ExactCQMSolver())
        self.assertTrue(node_partitions == {})

    def test_typical_cases(self):

        G = nx.complete_graph(8)

        node_partitions = dnx.partition(G, num_partitions=4, sampler=ExactCQMSolver())
        for i in range(4):
            self.assertTrue(sum(x == i for x in node_partitions.values()) == 2) # 4 equally sized subsets


        G = nx.complete_graph(10)
        node_partitions = dnx.partition(G, sampler=ExactCQMSolver())
        self.assertTrue(sum(x == 0 for x in node_partitions.values()) == 5)  # half of the nodes in subset '0'


        nx.set_edge_attributes(G, 1, 'weight')
        node_partitions = dnx.partition(G, sampler=ExactCQMSolver())
        self.assertTrue(sum(x == 0 for x in node_partitions.values()) == 5)  # half of the nodes in subset '0'


        G = nx.Graph()
        G.add_edges_from([(0, 1), (0, 2), (1, 2), (1, 3), (3, 4), (2, 4)])
        node_partitions = dnx.partition(G, sampler=ExactCQMSolver())
        self.assertTrue(sum(x == 0 for x in node_partitions.values()) in (2, 3)) # either 2 or 3 nodes in subset '0' (ditto '1')
        
        G = nx.Graph()
        G.add_edges_from([(0, 1), (0, 2), (1, 2), (2, 3), (3, 4), (3, 5), (4, 5)])
        node_partitions = dnx.partition(G, sampler=ExactCQMSolver())
        self.assertTrue(node_partitions[0] == node_partitions[1] == node_partitions[2])
        self.assertTrue(node_partitions[3] == node_partitions[4] == node_partitions[5])
        
        
        nx.set_edge_attributes(G, values = 1, name = 'weight')
        nx.set_edge_attributes(G, values = {(2, 3): 100}, name='weight')
        node_partitions = dnx.partition(G, sampler=ExactCQMSolver())
        self.assertTrue(node_partitions[2] == node_partitions[3]) # weight edges are respected


# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dwave_networkx.algorithms.independent_set import maximum_weighted_independent_set
from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ['min_weighted_vertex_cover', 'min_vertex_cover', 'is_vertex_cover']


@binary_quadratic_model_sampler(2)
def min_weighted_vertex_cover(G, weight=None, sampler=None, lagrange=2.0, **sampler_args):
    """Returns an approximate minimum weighted vertex cover.

    Defines a QUBO with ground states corresponding to a minimum weighted
    vertex cover and uses the sampler to sample from it.

    A vertex cover is a set of vertices such that each edge of the graph
    is incident with at least one vertex in the set. A minimum weighted
    vertex cover is the vertex cover of minimum total node weight.

    Parameters
    ----------
    G : NetworkX graph

    weight : string, optional (default None)
        If None, every node has equal weight. If a string, use this node
        attribute as the node weight. A node without this attribute is
        assumed to have max weight.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    lagrange : optional (default 2)
        Lagrange parameter to weight constraints versus objective.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    vertex_cover : list
       List of nodes that the form a the minimum weighted vertex cover, as
       determined by the given sampler.

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    https://en.wikipedia.org/wiki/Vertex_cover

    https://en.wikipedia.org/wiki/Quadratic_unconstrained_binary_optimization

    References
    ----------
    Based on the formulation presented in [AL]_

    """
    indep_nodes = set(maximum_weighted_independent_set(G, weight, sampler, lagrange, **sampler_args))
    return [v for v in G if v not in indep_nodes]


@binary_quadratic_model_sampler(1)
def min_vertex_cover(G, sampler=None, lagrange=2.0, **sampler_args):
    """Returns an approximate minimum vertex cover.

    Defines a QUBO with ground states corresponding to a minimum
    vertex cover and uses the sampler to sample from it.

    A vertex cover is a set of vertices such that each edge of the graph
    is incident with at least one vertex in the set. A minimum vertex cover
    is the vertex cover of smallest size.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum vertex cover.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    lagrange : optional (default 2)
        Lagrange parameter to weight constraints versus objective.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    vertex_cover : list
       List of nodes that form a minimum vertex cover, as
       determined by the given sampler.

    Examples
    --------
    This example uses a sampler from
    `dimod <https://github.com/dwavesystems/dimod>`_ to find a minimum vertex
    cover for a Chimera unit cell. Both the horizontal (vertices 0,1,2,3) and
    vertical (vertices 4,5,6,7) tiles connect to all 16 edges, so repeated
    executions can return either set.

    >>> import dwave_networkx as dnx
    >>> import dimod
    >>> sampler = dimod.ExactSolver()  # small testing sampler
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> G.remove_node(7)  # to give a unique solution
    >>> dnx.min_vertex_cover(G, sampler, lagrange=2.0)
    [4, 5, 6]

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------
    https://en.wikipedia.org/wiki/Vertex_cover

    https://en.wikipedia.org/wiki/Quadratic_unconstrained_binary_optimization

    .. [AL] Lucas, A. (2014). Ising formulations of many NP problems.
       Frontiers in Physics, Volume 2, Article 5.

    """
    return min_weighted_vertex_cover(G, None, sampler, lagrange, **sampler_args)


def is_vertex_cover(G, vertex_cover):
    """Determines whether the given set of vertices is a vertex cover of graph G.

    A vertex cover is a set of vertices such that each edge of the graph
    is incident with at least one vertex in the set.

    Parameters
    ----------
    G : NetworkX graph
       The graph on which to check the vertex cover.

    vertex_cover :
       Iterable of nodes.

    Returns
    -------
    is_cover : bool
        True if the given iterable forms a vertex cover.

    Examples
    --------
    This example checks two covers for a graph, G, of a single Chimera
    unit cell. The first uses the set of the four horizontal qubits, which
    do constitute a cover; the second set removes one node.

    >>> import dwave_networkx as dnx
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> cover = [0, 1, 2, 3]
    >>> dnx.is_vertex_cover(G,cover)
    True
    >>> cover = [0, 1, 2]
    >>> dnx.is_vertex_cover(G,cover)
    False

    """
    cover = set(vertex_cover)
    return all(u in cover or v in cover for u, v in G.edges)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import dwave_networkx as dnx


class Test_markov_network(unittest.TestCase):
    def test_empty(self):
        MN = dnx.markov_network({})

    def test_one_node(self):
        MN = dnx.markov_network({'a': {(0,): 1.2, (1,): .4}})

    def test_one_edge(self):
        MN = dnx.markov_network({'ab': {(0, 0): 1.2, (1, 0): .4, (0, 1): 1.3, (1, 1): -4}})



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import dwave_networkx as dnx
from dwave_networkx.utils import binary_quadratic_model_sampler


class MockSampler:
    def sample_ising(self, h, J):
        pass

    def sample_qubo(self, Q):
        pass


@binary_quadratic_model_sampler(1)
def mock_function(G, sampler=None, **sampler_args):
    assert sampler is not None


class TestDecorators(unittest.TestCase):

    def test_default_set(self):
        dnx.set_default_sampler(MockSampler())
        mock_function(0)
        dnx.unset_default_sampler()

        self.assertEqual(dnx.get_default_sampler(), None,
                         "sampler did not unset correctly")

    def test_no_sampler_set(self):
        with self.assertRaises(dnx.DWaveNetworkXMissingSampler):
            mock_function(0)

    def test_sampler_provided(self):
        mock_function(0, MockSampler())



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools

from collections import defaultdict

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ["traveling_salesperson",
           "traveling_salesperson_qubo",
           "traveling_salesman",
           "traveling_salesman_qubo",
           "is_hamiltonian_path",
           ]


@binary_quadratic_model_sampler(1)
def traveling_salesperson(G, sampler=None, lagrange=None, weight='weight',
                          start=None, **sampler_args):
    """Returns an approximate minimum traveling salesperson route.

    Defines a QUBO with ground states corresponding to the
    minimum routes and uses the sampler to sample
    from it.

    A route is a cycle in the graph that reaches each node exactly once.
    A minimum route is a route with the smallest total edge weight.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum traveling salesperson route.
        This should be a complete graph with non-zero weights on every edge.

    sampler :
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    lagrange : number, optional (default None)
        Lagrange parameter to weight constraints (visit every city once)
        versus objective (shortest distance route).

    weight : optional (default 'weight')
        The name of the edge attribute containing the weight.

    start : node, optional
        If provided, the route will begin at `start`.

    sampler_args :
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    route : list
       List of nodes in order to be visited on a route

    Examples
    --------

    >>> import dimod
    ...
    >>> G = nx.Graph()
    >>> G.add_weighted_edges_from({(0, 1, .1), (0, 2, .5), (0, 3, .1), (1, 2, .1),
    ...                            (1, 3, .5), (2, 3, .1)})
    >>> dnx.traveling_salesperson(G, dimod.ExactSolver(), start=0) # doctest: +SKIP
    [0, 1, 2, 3]

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    """
    # Get a QUBO representation of the problem
    Q = traveling_salesperson_qubo(G, lagrange, weight)

    # use the sampler to find low energy states
    response = sampler.sample_qubo(Q, **sampler_args)

    sample = response.first.sample

    route = [None]*len(G)
    for (city, time), val in sample.items():
        if val:
            route[time] = city

    if start is not None and route[0] != start:
        # rotate to put the start in front
        idx = route.index(start)
        route = route[idx:] + route[:idx]

    return route


traveling_salesman = traveling_salesperson


def traveling_salesperson_qubo(G, lagrange=None, weight='weight', missing_edge_weight=None):
    """Return the QUBO with ground states corresponding to a minimum TSP route.

    If :math:`|G|` is the number of nodes in the graph, the resulting qubo will have:

    * :math:`|G|^2` variables/nodes
    * :math:`2 |G|^2 (|G| - 1)` interactions/edges

    Parameters
    ----------
    G : NetworkX graph
        A complete graph in which each edge has a attribute giving its weight.

    lagrange : number, optional (default None)
        Lagrange parameter to weight constraints (no edges within set)
        versus objective (largest set possible).

    weight : optional (default 'weight')
        The name of the edge attribute containing the weight.
    
    missing_edge_weight : number, optional (default None)
        For bi-directional graphs, the weight given to missing edges.
        If None is given (the default), missing edges will be set to
        the sum of all weights.

    Returns
    -------
    QUBO : dict
       The QUBO with ground states corresponding to a minimum travelling
       salesperson route. The QUBO variables are labelled `(c, t)` where `c`
       is a node in `G` and `t` is the time index. For instance, if `('a', 0)`
       is 1 in the ground state, that means the node 'a' is visted first.

    """
    N = G.number_of_nodes()

    if lagrange is None:
        # If no lagrange parameter provided, set to 'average' tour length.
        # Usually a good estimate for a lagrange parameter is between 75-150%
        # of the objective function value, so we come up with an estimate for 
        # tour length and use that.
        if G.number_of_edges()>0:
            lagrange = G.size(weight=weight)*G.number_of_nodes()/G.number_of_edges()
        else:
            lagrange = 2
    
    # calculate default missing_edge_weight if required
    if missing_edge_weight is None:
        # networkx method to calculate sum of all weights
        missing_edge_weight = G.size(weight=weight)

    # some input checking
    if N in (1, 2):
        msg = "graph must have at least 3 nodes or be empty"
        raise ValueError(msg)

    # Creating the QUBO
    Q = defaultdict(float)

    # Constraint that each row has exactly one 1
    for node in G:
        for pos_1 in range(N):
            Q[((node, pos_1), (node, pos_1))] -= lagrange
            for pos_2 in range(pos_1+1, N):
                Q[((node, pos_1), (node, pos_2))] += 2.0*lagrange

    # Constraint that each col has exactly one 1
    for pos in range(N):
        for node_1 in G:
            Q[((node_1, pos), (node_1, pos))] -= lagrange
            for node_2 in set(G)-{node_1}:
                # QUBO coefficient is 2*lagrange, but we are placing this value 
                # above *and* below the diagonal, so we put half in each position.
                Q[((node_1, pos), (node_2, pos))] += lagrange

    # Objective that minimizes distance
    for u, v in itertools.combinations(G.nodes, 2):
        for pos in range(N):
            nextpos = (pos + 1) % N

            # going from u -> v
            try:
                value = G[u][v][weight]
            except KeyError:
                value = missing_edge_weight

            Q[((u, pos), (v, nextpos))] += value

            # going from v -> u
            try:
                value = G[v][u][weight]
            except KeyError:
                value = missing_edge_weight

            Q[((v, pos), (u, nextpos))] += value

    return Q


traveling_salesman_qubo = traveling_salesperson_qubo


def is_hamiltonian_path(G, route):
    """Determines whether the given list forms a valid TSP route.

    A travelling salesperson route must visit each city exactly once.

    Parameters
    ----------
    G : NetworkX graph

        The graph on which to check the route.

    route : list

        List of nodes in the order that they are visited.

    Returns
    -------
    is_valid : bool
        True if route forms a valid travelling salesperson route.

    """

    return (set(route) == set(G))



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import networkx as nx 
import dwave_networkx as dnx 
import dimod

# Use basic simulated annealer
sampler = dimod.SimulatedAnnealingSampler()

# Set up a Networkx Graph
G = nx.Graph()
G.add_edges_from([(1,2),(1,3),(2,3),(3,4),(3,5),(4,5),(4,6),(5,6),(6,7)])

# Find the maximum independent set, which is known in this case to be of length 3
candidate = dnx.maximum_independent_set(G, sampler)
if dnx.is_independent_set(G, candidate) and len(candidate) == 3:
    print(candidate, " is a maximum independent set")
else:
    print(candidate, " is not a minimum vertex coloring")


# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dwave_networkx.exceptions import DWaveNetworkXException
from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ["maximum_cut", "weighted_maximum_cut"]


@binary_quadratic_model_sampler(1)
def maximum_cut(G, sampler=None, **sampler_args):
    """Returns an approximate maximum cut.

    Defines an Ising problem with ground states corresponding to
    a maximum cut and uses the sampler to sample from it.

    A maximum cut is a subset S of the vertices of G such that
    the number of edges between S and the complementary subset
    is as large as possible.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximum cut.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    S : set
        A maximum cut of G.

    Example
    -------
    This example uses a sampler from
    `dimod <https://github.com/dwavesystems/dimod>`_ to find a maximum cut
    for a graph of a Chimera unit cell created using the `chimera_graph()`
    function.

    >>> import dimod
    ...
    >>> sampler = dimod.SimulatedAnnealingSampler()
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> cut = dnx.maximum_cut(G, sampler)

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    """
    # In order to form the Ising problem, we want to increase the
    # energy by 1 for each edge between two nodes of the same color.
    # The linear biases can all be 0.
    h = {v: 0. for v in G}
    J = {(u, v): 1 for u, v in G.edges}

    # draw the lowest energy sample from the sampler
    response = sampler.sample_ising(h, J, **sampler_args)
    sample = next(iter(response))

    return set(v for v in G if sample[v] >= 0)


@binary_quadratic_model_sampler(1)
def weighted_maximum_cut(G, sampler=None, **sampler_args):
    """Returns an approximate weighted maximum cut.

    Defines an Ising problem with ground states corresponding to
    a weighted maximum cut and uses the sampler to sample from it.

    A weighted maximum cut is a subset S of the vertices of G that
    maximizes the sum of the edge weights between S and its
    complementary subset.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a weighted maximum cut. Each edge in G should
        have a numeric `weight` attribute.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    S : set
        A maximum cut of G.

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    """
    # In order to form the Ising problem, we want to increase the
    # energy by 1 for each edge between two nodes of the same color.
    # The linear biases can all be 0.
    h = {v: 0. for v in G}
    try:
        J = {(u, v): G[u][v]['weight'] for u, v in G.edges}
    except KeyError:
        raise DWaveNetworkXException("edges must have 'weight' attribute")

    # draw the lowest energy sample from the sampler
    response = sampler.sample_ising(h, J, **sampler_args)
    sample = next(iter(response))

    return set(v for v in G if sample[v] >= 0)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools
import numbers
import warnings

import dimod
import networkx as nx

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ['is_matching',
           'is_maximal_matching',
           'matching_bqm',
           'maximal_matching_bqm',
           'min_maximal_matching',
           'min_maximal_matching_bqm',
           ]


def matching_bqm(G):
    """Find a binary quadratic model for the graph's matchings.

    A matching is a subset of edges in which no node occurs more than
    once. This function returns a binary quadratic model (BQM) with ground
    states corresponding to the possible matchings of G.

    Finding valid matchings can be done in polynomial time, so finding matching
    with BQMs is generally inefficient.
    This BQM may be useful when combined with other constraints and objectives.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a matching.

    Returns
    -------
    bqm : :class:`dimod.BinaryQuadraticModel`
        A binary quadratic model with ground states corresponding to a
        matching. The variables of the BQM are the edges of `G` as frozensets.
        The BQM's ground state energy is 0 by construction.
        The energy of the first excited state is 1.

    """
    bqm = dimod.BinaryQuadraticModel.empty('BINARY')

    # add the edges of G as variables
    for edge in G.edges:
        bqm.add_variable(frozenset(edge), 0)

    for node in G:
        for edge0, edge1 in itertools.combinations(G.edges(node), 2):
            u = frozenset(edge0)
            v = frozenset(edge1)
            bqm.add_interaction(u, v, 1)

    return bqm


def maximal_matching_bqm(G, lagrange=None):
    """Find a binary quadratic model for the graph's maximal matchings.

    A matching is a subset of edges in which no node occurs more than
    once. A maximal matching is one in which no edges from G can be
    added without violating the matching rule.
    This function returns a binary quadratic model (BQM) with ground
    states corresponding to the possible maximal matchings of G.

    Finding maximal matchings can be done in polynomial time, so finding
    maximal matching with BQMs is generally inefficient.
    This BQM may be useful when combined with other constraints and objectives.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximal matching.

    lagrange : float (optional)
        The Lagrange multiplier for the matching constraint. Should be positive
        and greater than `max_degree - 2`.
        Defaults to `1.25 * (max_degree - 2)`.

    Returns
    -------
    bqm : :class:`dimod.BinaryQuadraticModel`
        A binary quadratic model with ground states corresponding to a maximal
        matching. The variables of the BQM are the edges of `G` as frozensets.
        The BQM's ground state energy is 0 by construction.

    """
    bqm = matching_bqm(G)

    if lagrange is None:
        delta = max((G.degree[v] for v in G), default=0)
        lagrange = max(1.25 * (delta - 2), 1)

    bqm.scale(lagrange)

    for node0, node1 in G.edges:
        # (1 - y_v - y_u + y_v*y_u) <- see paper

        bqm.offset += 1

        for edge in G.edges(node0):
            bqm.linear[frozenset(edge)] -= 1

        for edge in G.edges(node1):
            bqm.linear[frozenset(edge)] -= 1

        for edge0 in G.edges(node0):
            u = frozenset(edge0)
            for edge1 in G.edges(node1):
                v = frozenset(edge1)
                if u == v:
                    bqm.linear[u] += 1
                else:
                    bqm.add_interaction(u, v, 1)

    return bqm


def min_maximal_matching_bqm(G, maximal_lagrange=2, matching_lagrange=None):
    """Find a binary quadratic model for the graph's minimum maximal matchings.

    A matching is a subset of edges in which no node occurs more than
    once. A maximal matching is one in which no edges from G can be
    added without violating the matching rule. A minimum maximal matching
    is a maximal matching that contains the smallest possible number of edges.
    This function returns a binary quadratic model (BQM) with ground
    states corresponding to the possible maximal matchings of G.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum maximal matching.

    maximal_lagrange : float (optional, default=2)
        The Lagrange multiplier for the maximal constraint. Should be greater
        than 1.

    matching_lagrange : float (optional)
        The Lagrange multiplier for the matching constraint. Should be positive
        and greater than `maximal_lagrange * max_degree - 2`.
        Defaults to `1.25 * (maximal_lagrange * max_degree - 2)`.

    Returns
    -------
    bqm : :class:`dimod.BinaryQuadraticModel`
        A binary quadratic model with ground states corresponding to a
        minimum maximal matching. The variables of the BQM are the edges
        of `G` as frozensets.

    """

    if matching_lagrange is not None:
        # we're going to scale the bqm by maximal_matching so undo that
        # for maximal_lagrange
        matching_lagrange /= maximal_lagrange
    bqm = maximal_matching_bqm(G, lagrange=matching_lagrange)
    bqm.scale(maximal_lagrange)

    for v in bqm.variables:
        bqm.linear[v] += 1

    return bqm


@binary_quadratic_model_sampler(1)
def maximal_matching(G, sampler=None, **sampler_args):
    """Finds an approximate maximal matching.

    Defines a QUBO with ground states corresponding to a maximal
    matching and uses the sampler to sample from it.

    A matching is a subset of edges in which no node occurs more than
    once. A maximal matching is one in which no edges from G can be
    added without violating the matching rule.

    Finding maximal matchings can be done is polynomial time, so this method
    is only useful pedagogically.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximal matching.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    matching : set
        A maximal matching of the graph.

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Matching on Wikipedia <https://w.wiki/r9s>`_

    `QUBO on Wikipedia <https://w.wiki/r9t>`_

    Based on the formulation presented in [AL]_

    """
    if not G.edges:
        return set()

    bqm = maximal_matching_bqm(G)
    sampleset = sampler.sample(bqm, **sampler_args)
    sample = sampleset.first.sample

    # the matching are the edges that are 1 in the sample
    return set(tuple(edge) for edge, val in sample.items() if val > 0)


@binary_quadratic_model_sampler(1)
def min_maximal_matching(G, sampler=None, **sampler_args):
    """Returns an approximate minimum maximal matching.

    Defines a QUBO with ground states corresponding to a minimum
    maximal matching and uses the sampler to sample from it.

    A matching is a subset of edges in which no node occurs more than
    once. A maximal matching is one in which no edges from G can be
    added without violating the matching rule. A minimum maximal
    matching is the smallest maximal matching for G.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum maximal matching.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    matching : set
        A minimum maximal matching of the graph.

    Example
    -------
    This example uses a sampler from
    `dimod <https://github.com/dwavesystems/dimod>`_ to find a minimum maximal
    matching for a Chimera unit cell.

    >>> import dimod
    >>> sampler = dimod.ExactSolver()
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> matching = dnx.min_maximal_matching(G, sampler)

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Matching on Wikipedia <https://w.wiki/r9s>`_

    `QUBO on Wikipedia <https://w.wiki/r9t>`_

    .. [AL] Lucas, A. (2014). Ising formulations of many NP problems.
       Frontiers in Physics, Volume 2, Article 5.

    """
    if not G.edges:
        return set()

    bqm = min_maximal_matching_bqm(G)
    sampleset = sampler.sample(bqm, **sampler_args)
    sample = sampleset.first.sample

    # the matching are the edges that are 1 in the sample
    return set(tuple(edge) for edge, val in sample.items() if val > 0)


def is_matching(edges):
    """Determine whether the given set of edges is a matching.

    Deprecated in favour of :func:`networkx.is_matching`.
    """
    warnings.warn("This method is deprecated, please use NetworkX's"
                  "nx.is_matching(G, edges) rather than dwave-networkx's "
                  "dnx.is_matching(edges)", DeprecationWarning, stacklevel=2)
    return len(set().union(*edges)) == len(edges) * 2


def is_maximal_matching(G, matching):
    """Determine whether the given set of edges is a maximal matching.

    Deprecated in favour of :func:`networkx.is_matching`.
    """
    warnings.warn("This method is deprecated, please use NetworkX's"
                  "nx.is_maximal_matching(G, edges) rather than "
                  "dwave-networkx's dnx.is_maximal_matching(G, edges)",
                  DeprecationWarning, stacklevel=2)
    touched_nodes = set().union(*matching)

    # first check if a matching
    if len(touched_nodes) != len(matching) * 2:
        return False

    # now for each edge, check that at least one of its variables is
    # already in the matching
    for (u, v) in G.edges:
        if u not in touched_nodes and v not in touched_nodes:
            return False

    return True



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Tools to visualize Pegasus lattices and weighted graph problems on them.
"""

import networkx as nx
from networkx import draw

from dwave_networkx.drawing.qubit_layout import draw_qubit_graph, draw_embedding, draw_yield
from dwave_networkx.generators.pegasus import pegasus_graph, pegasus_coordinates
from dwave_networkx.drawing.chimera_layout import chimera_node_placer_2d


__all__ = ['pegasus_layout',
           'draw_pegasus',
           'draw_pegasus_embedding',
           'draw_pegasus_yield',
           ]


def pegasus_layout(G, scale=1., center=None, dim=2, crosses=False):
    """Positions the nodes of graph G in a Pegasus topology.

    `NumPy <https://scipy.org>`_ is required for this function.

    Parameters
    ----------
    G : NetworkX graph
        A Pegasus graph or a subgraph of a Pegasus graph, as produced by
        the :func:`dwave_networkx.pegasus_graph` function.

    scale : float (default 1.)
        Scale factor. A setting of ``scale = 1`` fits all positions within
        [0, 1] on the x-axis and [-1, 0] on the y-axis.

    center : None or array (default None)
        Coordinates of the top left corner.

    dim : int (default 2)
        Number of dimensions. When dim > 2, all extra dimensions are
        set to 0.

    crosses: boolean (optional, default False)
        If True, :math:`K_{4,4}` subgraphs are shown in a cross
        rather than L configuration. Ignored if G is defined with
        ``nice_coordinates=True``.

    Returns
    -------
    pos : dict
        Positions as a dictionary keyed by node.

    Examples
    --------
        This example gives the positions of a Pegasus lattice of size 2.

    >>> G = dnx.pegasus_graph(2)
    >>> pos = dnx.pegasus_layout(G)

    """

    if not isinstance(G, nx.Graph) or G.graph.get("family") != "pegasus":
        raise ValueError("G must be generated by dwave_networkx.pegasus_graph")

    if G.graph.get('labels') == 'nice':
        m = 3*(G.graph['rows']-1)
        c_coords = chimera_node_placer_2d(m, m, 4, scale=scale, center=center, dim=dim)
        def xy_coords(t, y, x, u, k):
            return c_coords(3*y+2-t, 3*x+t, u, k)
        pos = {v: xy_coords(*v) for v in G.nodes()}
    else:
        xy_coords = pegasus_node_placer_2d(G, scale, center, dim, crosses=crosses)

        if G.graph.get('labels') == 'coordinate':
            pos = {v: xy_coords(*v) for v in G.nodes()}
        elif G.graph.get('data'):
            pos = {v: xy_coords(*dat['pegasus_index']) for v, dat in G.nodes(data=True)}
        else:
            m = G.graph.get('rows')
            coord = pegasus_coordinates(m)
            pos = {v: xy_coords(*coord.linear_to_pegasus(v)) for v in G.nodes()}

    return pos


def pegasus_node_placer_2d(G, scale=1., center=None, dim=2, crosses=False):
    """Generates a function to convert Pegasus indices to plottable coordinates.

    Parameters
    ----------
    G : NetworkX graph
        A Pegasus graph or a subgraph of a Pegasus graph, as produced by
        the :func:`dwave_networkx.pegasus_graph` function.

    scale : float (default 1.)
        Scale factor. A setting of ``scale = 1`` fits all positions within
        [0, 1] on the x-axis and [-1, 0] on the y-axis.

    center : None or array (default None)
        Coordinates of the top left corner.

    dim : int (default 2)
        Number of dimensions. When dim > 2, all extra dimensions are
        set to 0.

    crosses: boolean (optional, default False)
        If True, :math:`K_{4,4}` subgraphs are shown in a cross
        rather than L configuration.

    Returns
    -------
    xy_coords : function
        A function that maps a Pegasus index (u, w, k, z) in a
        Pegasus lattice to plottable x,y coordinates.

    """
    import numpy as np

    m = G.graph.get('rows')
    h_offsets = G.graph.get("horizontal_offsets")
    v_offsets = G.graph.get("vertical_offsets")
    tile_width = G.graph.get("tile")
    tile_center = tile_width / 2 - .5

    # want the enter plot to fill in [0, 1] when scale=1
    scale /= m * tile_width

    if center is None:
        center = np.zeros(dim)
    else:
        center = np.asarray(center)

    paddims = dim - 2
    if paddims < 0:
        raise ValueError("layout must have at least two dimensions")

    if len(center) != dim:
        raise ValueError("length of center coordinates must match dimension of layout")

    if crosses:
        # adjustment for crosses
        cross_shift = 2.
    else:
        cross_shift = 0.

    def _xy_coords(u, w, k, z):
        # orientation, major perpendicular offset, minor perpendicular offset, parallel offset

        if k % 2:
            p = -.1
        else:
            p = .1

        if u:
            xy = np.array([z*tile_width+h_offsets[k] + tile_center, -tile_width*w-k-p+cross_shift])
        else:
            xy = np.array([tile_width*w+k+p+cross_shift, -z*tile_width-v_offsets[k]-tile_center])

        # convention for Pegasus-lattice pictures is to invert the y-axis
        return np.hstack((xy * scale, np.zeros(paddims))) + center

    return _xy_coords


def draw_pegasus(G, crosses=False, **kwargs):
    """Draws graph G in a Pegasus topology.

    If ``linear_biases`` and/or ``quadratic_biases`` are provided, these
    are visualized on the plot.

    Parameters
    ----------
    G : NetworkX graph
        A Pegasus graph or a subgraph of a Pegasus graph, as produced by
        the :func:`dwave_networkx.pegasus_graph` function.

    linear_biases : dict (optional, default {})
        Biases as a dict, of form {node: bias, ...}, where keys are
        nodes in G and biases are numeric.

    quadratic_biases : dict (optional, default {})
        Biases as a dict, of form {edge: bias, ...}, where keys are
        edges in G and biases are numeric. Self-loop
        edges (i.e., :math:`i=j`) are treated as linear biases.

    crosses: boolean (optional, default False)
        If True, :math:`K_{4,4}` subgraphs are shown in a cross
        rather than L configuration. Ignored if G is defined with
        ``nice_coordinates=True``.

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the ``pos`` parameter, which is not used by this
       function. If ``linear_biases`` or ``quadratic_biases`` are provided,
       any provided ``node_color`` or ``edge_color`` arguments are ignored.

    Examples
    --------
        This example plots a Pegasus graph with size parameter 2.

    >>> import networkx as nx
    >>> import dwave_networkx as dnx
    >>> import matplotlib.pyplot as plt   # doctest: +SKIP
    >>> G = dnx.pegasus_graph(2)
    >>> dnx.draw_pegasus(G)    # doctest: +SKIP
    >>> plt.show()    # doctest: +SKIP

    """

    draw_qubit_graph(G, pegasus_layout(G, crosses=crosses), **kwargs)


def draw_pegasus_embedding(G, *args, **kwargs):
    """Draws an embedding onto Pegasus graph G.

    Parameters
    ----------
    G : NetworkX graph
        A Pegasus graph or a subgraph of a Pegasus graph, as produced by
        the :func:`dwave_networkx.pegasus_graph` function.

    emb : dict
        Chains, as a dict of form {qubit: chain, ...}, where qubits are
        nodes in G and chains are iterables of qubit labels.

    embedded_graph : NetworkX graph (optional, default None)
        A graph that contains all keys of ``emb`` as nodes.  If specified,
        edges of G are considered interactions if and only if (1) they
        exist between two chains of ``emb`` and (2) their keys are connected
        by an edge in this graph. If given, only couplers between chains
        based on this graph are displayed.

    interaction_edges : list (optional, default None)
        A list of edges used as interactions. If given,
        only these couplers are displayed.

    show_labels: boolean (optional, default False)
        If True, each chain in ``emb`` is labelled with its key.

    chain_color : dict (optional, default None)
        Colors as a dict of form {node: rgba_color, ...} associated with
        each key in ``emb``, where colors are length-4 tuples of floats
        between 0 and 1 inclusive. If None, each chain is assigned a
        different color.

    unused_color : tuple (optional, default (0.9,0.9,0.9,1.0))
        Color for nodes of G that are not part of chains, and edges
        that are neither chain edges nor interactions. If None, these
        nodes and edges are not shown.

    crosses: boolean (optional, default False)
        If True, :math:`K_{4,4}` subgraphs are shown in a cross
        rather than L configuration. Ignored if G is defined with
        ``nice_coordinates=True``.

    overlapped_embedding: boolean (optional, default False)
        If True, chains in ``emb`` may overlap (contain the same vertices
        in G), and these overlaps are displayed as concentric circles.

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the ``pos`` parameter, which is not used by this
       function. If ``linear_biases`` or ``quadratic_biases`` are provided,
       any provided ``node_color`` or ``edge_color`` arguments are ignored.
    """
    crosses = kwargs.pop("crosses", False)
    draw_embedding(G, pegasus_layout(G, crosses=crosses), *args, **kwargs)

def draw_pegasus_yield(G, **kwargs):
    """Draws the given graph G with highlighted faults, according to layout.

    Parameters
    ----------
    G : NetworkX graph
        Graph to be parsed for faults.

    unused_color : tuple or color string (optional, default (0.9,0.9,0.9,1.0))
        The color to use for nodes and edges of G which are not faults.
        If unused_color is None, these nodes and edges will not be shown at all.

    fault_color : tuple or color string (optional, default (1.0,0.0,0.0,1.0))
        A color to represent nodes absent from the graph G. Colors should be
        length-4 tuples of floats between 0 and 1 inclusive.

    fault_shape : string, optional (default='x')
        The shape of the fault nodes. Specification is as matplotlib.scatter
        marker, one of 'so^>v<dph8'.

    fault_style : string, optional (default='dashed')
        Edge fault line style (solid|dashed|dotted|dashdot)

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the `pos` parameter which is not used by this
       function. If `linear_biases` or `quadratic_biases` are provided,
       any provided `node_color` or `edge_color` arguments are ignored.
    """
    try:
        assert(G.graph["family"] == "pegasus")
        m = G.graph['columns']
        offset_lists = (G.graph['vertical_offsets'], G.graph['horizontal_offsets'])
        coordinates = G.graph["labels"] == "coordinate"
        nice = G.graph["labels"] == "nice"
        # Can't interpret fabric_only from graph attributes
    except:
        raise ValueError("Target pegasus graph needs to have columns, rows, \
        tile, and label attributes to be able to identify faulty qubits.")


    perfect_graph = pegasus_graph(m, offset_lists=offset_lists, coordinates=coordinates, nice_coordinates=nice)

    draw_yield(G, pegasus_layout(perfect_graph), perfect_graph, **kwargs)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import networkx as nx
import dwave_networkx as dnx

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ["maximum_clique", "clique_number", "is_clique"]

@binary_quadratic_model_sampler(1)
def maximum_clique(G, sampler=None, lagrange=2.0, **sampler_args):
    """
    Returns an approximate maximum clique.
    A clique in an undirected graph, G = (V, E), is a subset of the vertex set
    :math:`C \subseteq V` such that for every two vertices in C there exists an edge
    connecting the two. This is equivalent to saying that the subgraph
    induced by C is complete (in some cases, the term clique may also refer
    to the subgraph). A maximum clique is a clique of the largest
    possible size in a given graph.

    This function works by finding the maximum independent set of the compliment
    graph of the given graph G which is equivalent to finding maximum clique.
    It defines a QUBO with ground states corresponding
    to a maximum weighted independent set and uses the sampler to sample from it.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximum clique.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    lagrange : optional (default 2)
        Lagrange parameter to weight constraints (no edges within set)
        versus objective (largest set possible).

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    clique_nodes : list
       List of nodes that form a maximum clique, as
       determined by the given sampler.

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Maximum Clique on Wikipedia <https://en.wikipedia.org/wiki/Clique_(graph_theory)>`_

    `Independent Set on Wikipedia <https://en.wikipedia.org/wiki/Independent_set_(graph_theory)>`_

    `QUBO on Wikipedia <https://en.wikipedia.org/wiki/Quadratic_unconstrained_binary_optimization>`_

    .. [AL] Lucas, A. (2014). Ising formulations of many NP problems.
       Frontiers in Physics, Volume 2, Article 5.
    """
    if G is None:
        raise ValueError("Expected NetworkX graph!")

    # finding the maximum clique in a graph is equivalent to finding
    # the independent set in the complementary graph
    complement_G = nx.complement(G)
    return dnx.maximum_independent_set(complement_G, sampler, lagrange, **sampler_args)


@binary_quadratic_model_sampler(1)
def clique_number(G, sampler=None, lagrange=2.0, **sampler_args):
    """
    Returns the number of vertices in the maximum clique of a graph.
    A maximum clique is a clique of the largest possible size in a given graph.
    The clique number math:`\omega(G)` of a graph G is the number of
    vertices in a maximum clique in G. The intersection number of
    G is the smallest number of cliques that together cover all edges of G.

    This function works by finding the maximum independent set of the compliment
    graph of the given graph G which is equivalent to finding maximum clique.
    It defines a QUBO with ground states corresponding
    to a maximum weighted independent set and uses the sampler to sample from it.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximum clique.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    lagrange : optional (default 2)
        Lagrange parameter to weight constraints (no edges within set)
        versus objective (largest set possible).

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    clique_nodes : list
       List of nodes that form a maximum clique, as
       determined by the given sampler.

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Maximum Clique on Wikipedia <https://en.wikipedia.org/wiki/Clique_(graph_theory)>`_
    """
    return len(maximum_clique(G, sampler, lagrange, **sampler_args))

def is_clique(G, clique_nodes):
    """Determines whether the given nodes form a clique.

    A clique is a subset of nodes of an undirected graph such that every two
    distinct nodes in the clique are adjacent.

    Parameters
    ----------
    G : NetworkX graph
       The graph on which to check the clique nodes.

    clique_nodes : list
       List of nodes that form a clique, as
       determined by the given sampler.

    Returns
    -------
    is_clique : bool
        True if clique_nodes forms a clique.

    Example
    -------
    This example checks two sets of nodes, both derived from a
    single Chimera unit cell, for an independent set. The first set is
    the horizontal tile's nodes; the second has nodes from the horizontal and
    verical tiles.

    >>> import dwave_networkx as dnx
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> dnx.is_clique(G, [0, 1, 2, 3])
    False
    >>> dnx.is_clique(G, [0, 4])
    True
    """
    for x in clique_nodes:
        for y in clique_nodes:
            if x != y:
                if not(G.has_edge(x,y)):
                    return False
    return True



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx
import dwave_networkx as dnx


class HeuristicCases:
    """Change the name for compatibility with nose."""

    def test_basic(self):
        G = nx.Graph()

        tw, order = self.heuristic(G)

        self.assertGreaterEqual(tw, 0)
        self.check_order(G, order)

    def test_complete(self):
        G = nx.complete_graph(10)

        tw, order = self.heuristic(G)
        self.assertGreaterEqual(tw, 9)
        self.check_order(G, order)

    def test_chimera(self):
        G = dnx.chimera_graph(2, 2, 4)

        tw, order = self.heuristic(G)
        self.assertGreaterEqual(tw, 8)
        self.check_order(G, order)

    def test_cycle(self):
        G = nx.cycle_graph(43)

        tw, order = self.heuristic(G)
        self.assertGreaterEqual(tw, 2)
        self.check_order(G, order)

    def test_grid(self):
        G = nx.grid_2d_graph(6, 7)

        tw, order = self.heuristic(G)
        self.assertGreaterEqual(tw, 6)
        self.check_order(G, order)

    def check_order(self, G, order):
        self.assertEqual(set(G), set(order))

    def test_disjoint(self):
        graph = nx.complete_graph(1)
        graph.add_node(1)

        tw, order = self.heuristic(graph)
        self.assertGreaterEqual(tw, 0)
        self.check_order(graph, order)

        graph = nx.complete_graph(4)
        graph.add_node(4)

        tw, order = self.heuristic(graph)
        self.assertGreaterEqual(tw, 3)
        self.check_order(graph, order)

        graph = nx.complete_graph(4)
        graph.add_edge(4, 5)

        tw, order = self.heuristic(graph)
        self.assertGreaterEqual(tw, 3)
        self.check_order(graph, order)

    def test_self_loop(self):
        graph = nx.complete_graph(3)
        graph.add_edge(0, 0)
        graph.add_edge(2, 2)

        tw, order = self.heuristic(graph)
        self.assertGreaterEqual(tw, 2)
        self.check_order(graph, order)


class TestMinWidth(unittest.TestCase, HeuristicCases):
    def setUp(self):
        self.heuristic = dnx.min_width_heuristic


class TestMinFill(unittest.TestCase, HeuristicCases):
    def setUp(self):
        self.heuristic = dnx.min_fill_heuristic


class TestMaxCardinality(unittest.TestCase, HeuristicCases):
    def setUp(self):
        self.heuristic = dnx.max_cardinality_heuristic


class TestMinorMinWidth(unittest.TestCase):
    def test_basic(self):
        G = nx.complete_graph(10)
        lb = dnx.minor_min_width(G)
        self.assertLessEqual(lb, len(G) - 1)
        self.assertEqual(len(G), 10)

    def test_disjoint(self):
        graph = nx.complete_graph(1)
        graph.add_node(1)

        lb = dnx.minor_min_width(graph)
        self.assertEqual(lb, 0)

        graph = nx.complete_graph(4)
        graph.add_node(4)

        lb = dnx.minor_min_width(graph)
        self.assertEqual(lb, 3)

        graph = nx.complete_graph(4)
        graph.add_edge(4, 5)

        lb = dnx.minor_min_width(graph)
        self.assertEqual(lb, 3)

    def test_self_loop(self):
        graph = nx.complete_graph(3)
        graph.add_edge(0, 0)
        graph.add_edge(2, 2)

        lb = dnx.minor_min_width(graph)


class TestSimplicialTests(unittest.TestCase):
    def test_typical(self):

        # every node in a complete graph is simplicial
        G = nx.complete_graph(100)
        for v in G:
            self.assertTrue(dnx.is_simplicial(G, v))

        # if we remove one edge then every node except the
        # two should be almost simplicial
        u = 0
        w = 1
        G.remove_edge(u, w)
        for v in G:
            if v not in (u, w):
                self.assertFalse(dnx.is_simplicial(G, v))
                self.assertTrue(dnx.is_almost_simplicial(G, v))
            else:
                self.assertTrue(dnx.is_simplicial(G, v))


class TestBranchAndBound(unittest.TestCase):
    def test_empty(self):
        G = nx.Graph()
        true_tw = 0

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

    def test_chimera(self):
        G = dnx.chimera_graph(1, 1, 4)
        true_tw = 4

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

        G = dnx.chimera_graph(2, 2, 3)
        true_tw = 6

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

        G = dnx.chimera_graph(1, 2, 4)
        true_tw = 4

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

        G = dnx.chimera_graph(2, 2, 4)
        true_tw = 8

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

    def test_complete(self):
        G = nx.complete_graph(100)
        true_tw = 99

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

    def test_complete_minus_1(self):
        G = nx.complete_graph(50)
        G.remove_edge(7, 6)
        true_tw = 48

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

    def test_cycle(self):
        G = nx.cycle_graph(167)
        true_tw = 2

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

    def test_grid(self):
        G = nx.grid_2d_graph(4, 6)
        true_tw = 4

        tw, order = dnx.treewidth_branch_and_bound(G)
        self.check_order(G, order)
        self.assertEqual(tw, true_tw)

    def check_order(self, G, order):
        self.assertEqual(set(G), set(order))

    def test_disjoint(self):
        graph = nx.complete_graph(1)
        graph.add_node(1)

        tw, order = dnx.treewidth_branch_and_bound(graph)
        self.check_order(graph, order)
        self.assertEqual(tw, 0)

        graph = nx.complete_graph(4)
        graph.add_node(4)

        tw, order = dnx.treewidth_branch_and_bound(graph)
        self.check_order(graph, order)
        self.assertEqual(tw, 3)

        graph = nx.complete_graph(4)
        graph.add_edge(4, 5)

        tw, order = dnx.treewidth_branch_and_bound(graph)
        self.check_order(graph, order)
        self.assertEqual(tw, 3)

    def test_upperbound_parameter(self):
        """We want to be able to give an upper bound to make execution faster."""
        graph = nx.complete_graph(3)

        # providing a treewidth without an order should still work, although
        # the order might not be found
        tw, order = dnx.treewidth_branch_and_bound(graph, treewidth_upperbound=2)
        self.assertEqual(len(order), 3)  # all nodes should be in the order

        tw, order = dnx.treewidth_branch_and_bound(graph, [0, 1, 2])
        self.assertEqual(len(order), 3)  # all nodes should be in the order

        # try with both
        tw, order = dnx.treewidth_branch_and_bound(graph, [0, 1, 2], 2)
        self.assertEqual(len(order), 3)  # all nodes should be in the order

    def test_incorrect_lowerbound(self):
        graph = nx.complete_graph(3)

        tw, order = dnx.treewidth_branch_and_bound(graph, treewidth_upperbound=1)
        self.assertEqual(order, [])  # no order produced

    def test_singleton(self):
        G = nx.Graph()
        G.add_node('a')

        tw, order = dnx.treewidth_branch_and_bound(G)

    def test_self_loop(self):
        graph = nx.complete_graph(3)
        graph.add_edge(0, 0)
        graph.add_edge(2, 2)

        tw, order = dnx.treewidth_branch_and_bound(graph)

        self.assertEqual(tw, 2)


class TestEliminationOrderWidth(unittest.TestCase):
    def test_trivial(self):
        G = nx.Graph()
        tw = dnx.elimination_order_width(G, [])
        self.assertEqual(tw, 0)

    def test_graphs(self):

        H = nx.complete_graph(2)
        H.add_edge(2, 3)

        graphs = [nx.complete_graph(7),
                  dnx.chimera_graph(2, 1, 3),
                  nx.balanced_tree(5, 3),
                  nx.barbell_graph(8, 11),
                  nx.cycle_graph(5),
                  H]

        for G in graphs:
            tw, order = dnx.treewidth_branch_and_bound(G)
            self.assertEqual(dnx.elimination_order_width(G, order), tw)

            tw, order = dnx.min_width_heuristic(G)
            self.assertEqual(dnx.elimination_order_width(G, order), tw)

            tw, order = dnx.min_fill_heuristic(G)
            self.assertEqual(dnx.elimination_order_width(G, order), tw)

            tw, order = dnx.max_cardinality_heuristic(G)
            self.assertEqual(dnx.elimination_order_width(G, order), tw)

    def test_exceptions(self):

        G = nx.complete_graph(6)
        order = range(4)

        with self.assertRaises(ValueError):
            dnx.elimination_order_width(G, order)

        order = range(7)
        with self.assertRaises(ValueError):
            dnx.elimination_order_width(G, order)


class TestChimeraEliminationOrder(unittest.TestCase):
    def test_variable_order(self):
        n = 8
        m = 10
        p = dnx.chimera_graph(n, m)
        o = dnx.chimera_elimination_order(n, m)
        tw = dnx.elimination_order_width(p, o)
        self.assertEqual(tw, 4*n)

        p = dnx.chimera_graph(m, n)
        o = dnx.chimera_elimination_order(m, n)
        tw = dnx.elimination_order_width(p, o)
        self.assertEqual(tw, 4*n)


class TestPegasusEliminationOrder(unittest.TestCase):
    def test_variable_order(self):
        n = 4
        p = dnx.pegasus_graph(n, fabric_only=False)
        o = dnx.pegasus_elimination_order(n)
        tw = dnx.elimination_order_width(p, o)
        self.assertEqual(tw, 12*n-4)

        p = dnx.pegasus_graph(n, fabric_only=False, coordinates=True)
        o = dnx.pegasus_elimination_order(n, coordinates=True)
        tw = dnx.elimination_order_width(p, o)
        self.assertEqual(tw, 12*n-4)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import dwave_networkx.generators
from dwave_networkx.generators import *

import dwave_networkx.algorithms
from dwave_networkx.algorithms import *

import dwave_networkx.utils
from dwave_networkx.exceptions import *

import dwave_networkx.default_sampler
from dwave_networkx.default_sampler import *

import dwave_networkx.drawing
from dwave_networkx.drawing import *

from dwave_networkx.package_info import __version__, __author__, \
    __authoremail__, __description__



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx
import dwave_networkx as dnx

from dimod import ExactSolver, SimulatedAnnealingSampler, qubo_energy


class TestMaxCut(unittest.TestCase):
    # def test_edge_cases(self):
    #     # get the empty graph
    #     G = nx.Graph()

    #     S = dnx.maximum_cut(G, ExactSolver())
    #     self.assertTrue(len(S) == 0)

    #     S = dnx.weighted_maximum_cut(G, ExactSolver())
    #     self.assertTrue(len(S) == 0)

    def test_typical_cases(self):

        G = nx.complete_graph(10)

        S = dnx.maximum_cut(G, ExactSolver())
        self.assertTrue(len(S) == 5)  # half of the nodes

        with self.assertRaises(dnx.DWaveNetworkXException):
            S = dnx.weighted_maximum_cut(G, ExactSolver())

        nx.set_edge_attributes(G, 1, 'weight')
        S = dnx.weighted_maximum_cut(G, ExactSolver())
        self.assertTrue(len(S) == 5)  # half of the nodes

        G = nx.Graph()
        G.add_edges_from([(0, 1), (0, 2), (1, 2), (1, 3), (3, 4), (2, 4)])
        S = dnx.maximum_cut(G, ExactSolver())
        self.assertTrue(len(S) in (2, 3))

        # this needs another one for weight



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import dimod

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ['sample_markov_network', 'markov_network_bqm']


###############################################################################
# The following code is partially based on https://github.com/tbabej/gibbs
#
# MIT License
# ===========
#
# Copyright 2017 Tomas Babej
# https://github.com/tbabej/gibbs
#
# This software is released under MIT licence.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


@binary_quadratic_model_sampler(1)
def sample_markov_network(MN, sampler=None, fixed_variables=None,
                          return_sampleset=False,
                          **sampler_args):
    """Samples from a markov network using the provided sampler.

    Parameters
    ----------
    G : NetworkX graph
        A Markov Network as returned by :func:`.markov_network`

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    fixed_variables : dict
        A dictionary of variable assignments to be fixed in the markov network.

    return_sampleset : bool (optional, default=False)
        If True, returns a :obj:`dimod.SampleSet` rather than a list of samples.

    **sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    samples : list[dict]/:obj:`dimod.SampleSet`
        A list of samples ordered from low-to-high energy or a sample set.

    Examples
    --------

    >>> import dimod
    ...
    >>> potentials = {('a', 'b'): {(0, 0): -1,
    ...                            (0, 1): .5,
    ...                            (1, 0): .5,
    ...                            (1, 1): 2}}
    >>> MN = dnx.markov_network(potentials)
    >>> sampler = dimod.ExactSolver()
    >>> samples = dnx.sample_markov_network(MN, sampler)
    >>> samples[0]     # doctest: +SKIP
    {'a': 0, 'b': 0}

    >>> import dimod
    ...
    >>> potentials = {('a', 'b'): {(0, 0): -1,
    ...                            (0, 1): .5,
    ...                            (1, 0): .5,
    ...                            (1, 1): 2}}
    >>> MN = dnx.markov_network(potentials)
    >>> sampler = dimod.ExactSolver()
    >>> samples = dnx.sample_markov_network(MN, sampler, return_sampleset=True)
    >>> samples.first       # doctest: +SKIP
    Sample(sample={'a': 0, 'b': 0}, energy=-1.0, num_occurrences=1)

    >>> import dimod
    ...
    >>> potentials = {('a', 'b'): {(0, 0): -1,
    ...                            (0, 1): .5,
    ...                            (1, 0): .5,
    ...                            (1, 1): 2},
    ...               ('b', 'c'): {(0, 0): -9,
    ...                            (0, 1): 1.2,
    ...                            (1, 0): 7.2,
    ...                            (1, 1): 5}}
    >>> MN = dnx.markov_network(potentials)
    >>> sampler = dimod.ExactSolver()
    >>> samples = dnx.sample_markov_network(MN, sampler, fixed_variables={'b': 0})
    >>> samples[0]           # doctest: +SKIP
    {'a': 0, 'c': 0, 'b': 0}

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    """

    bqm = markov_network_bqm(MN)

    if fixed_variables:
        # we can modify in-place since we just made it
        bqm.fix_variables(fixed_variables)

    sampleset = sampler.sample(bqm, **sampler_args)

    if fixed_variables:
        # add the variables back in
        sampleset = dimod.append_variables(sampleset, fixed_variables)

    if return_sampleset:
        return sampleset
    else:
        return list(map(dict, sampleset.samples()))


def markov_network_bqm(MN):
    """Construct a binary quadratic model for a markov network.


    Parameters
    ----------
    G : NetworkX graph
        A Markov Network as returned by :func:`.markov_network`

    Returns
    -------
    bqm : :obj:`dimod.BinaryQuadraticModel`
        A binary quadratic model.

    """

    bqm = dimod.BinaryQuadraticModel.empty(dimod.BINARY)

    # the variable potentials
    for v, ddict in MN.nodes(data=True, default=None):
        potential = ddict.get('potential', None)

        if potential is None:
            continue

        # for single nodes we don't need to worry about order

        phi0 = potential[(0,)]
        phi1 = potential[(1,)]

        bqm.add_variable(v, phi1 - phi0)
        bqm.offset += phi0

    # the interaction potentials
    for u, v, ddict in MN.edges(data=True, default=None):
        potential = ddict.get('potential', None)

        if potential is None:
            continue

        # in python<=3.5 the edge order might not be consistent so we use the
        # one that was stored
        order = ddict['order']
        u, v = order

        phi00 = potential[(0, 0)]
        phi01 = potential[(0, 1)]
        phi10 = potential[(1, 0)]
        phi11 = potential[(1, 1)]

        bqm.add_variable(u, phi10 - phi00)
        bqm.add_variable(v, phi01 - phi00)
        bqm.add_interaction(u, v, phi11 - phi10 - phi01 + phi00)
        bqm.offset += phi00

    return bqm



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools
import collections.abc as abc

from collections import namedtuple

import networkx as nx


__all__ = 'markov_network',


###############################################################################
# The following code is partially based on https://github.com/tbabej/gibbs
#
# MIT License
# ===========
#
# Copyright 2017 Tomas Babej
# https://github.com/tbabej/gibbs
#
# This software is released under MIT licence.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


def markov_network(potentials):
    """Creates a Markov Network from potentials.

    A Markov Network is also knows as a `Markov Random Field`_

    Parameters
    ----------
    potentials : dict[tuple, dict]
        A dict where the keys are either nodes or edges and the values are a
        dictionary of potentials. The potential dict should map each possible
        assignment of the nodes/edges to their energy.

    Returns
    -------
    MN : :obj:`networkx.Graph`
        A markov network as a graph where each node/edge stores its potential
        dict as above.

    Examples
    --------
    >>> potentials = {('a', 'b'): {(0, 0): -1,
    ...                            (0, 1): .5,
    ...                            (1, 0): .5,
    ...                            (1, 1): 2}}
    >>> MN = dnx.markov_network(potentials)
    >>> MN['a']['b']['potential'][(0, 0)]
    -1

    .. _Markov Random Field: https://en.wikipedia.org/wiki/Markov_random_field

    """

    G = nx.Graph()

    G.name = 'markov_network({!r})'.format(potentials)

    # we use 'clique' because the keys of potentials can be either nodes or
    # edges, but in either case they are fully connected.
    for clique, phis in potentials.items():

        num_vars = len(clique)

        # because this data potentially wont be used for a while, let's do some
        # input checking now and save some debugging issues later
        if not isinstance(phis, abc.Mapping):
            raise TypeError("phis should be a dict")
        elif not all(config in phis for config in itertools.product((0, 1), repeat=num_vars)):
            raise ValueError("not all potentials provided for {!r}".format(clique))

        if num_vars == 1:
            u, = clique
            G.add_node(u, potential=phis)
        elif num_vars == 2:
            u, v = clique
            # in python<=3.5 the edge order might not be consistent so we store
            # the relevant order of the variables relative to the potentials
            G.add_edge(u, v, potential=phis, order=(u, v))
        else:
            # developer note: in principle supporting larger cliques can be done
            # using higher-order, but it would make the use of networkx graphs
            # far more difficult
            raise ValueError("Only supports cliques up to size 2")

    return G



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Decorators allow for input checking and default parameter setting for
algorithms.
"""

import functools
import inspect

import dwave_networkx as dnx

__all__ = ['binary_quadratic_model_sampler']


def binary_quadratic_model_sampler(which_args):
    """Decorator to validate sampler arguments.

    Parameters
    ----------
    which_args : int or sequence of ints
        Location of the sampler arguments of the input function in the form
        `function_name(args, *kw)`. If more than one
        sampler is allowed, can be a list of locations.

    """
    def decorator(f):
        @functools.wraps(f)
        def func(*args, **kwargs):
            bound_arguments = inspect.signature(f).bind(*args, **kwargs)
            bound_arguments.apply_defaults()

            args = bound_arguments.args
            kw = bound_arguments.kwargs

            if isinstance(which_args, int):
                iter_args = (which_args,)
            else:
                iter_args = iter(which_args)

            # check each sampler for the correct methods
            new_args = [arg for arg in args]
            for idx in iter_args:
                sampler = args[idx]

                # if no sampler is provided, get the default sampler if it has
                # been set
                if sampler is None:
                    # this sampler has already been vetted
                    default_sampler = dnx.get_default_sampler()
                    if default_sampler is None:
                        raise dnx.DWaveNetworkXMissingSampler('no default sampler set')
                    new_args[idx] = default_sampler
                    continue

                if not hasattr(sampler, "sample_qubo") or not callable(sampler.sample_qubo):
                    raise TypeError("expected sampler to have a 'sample_qubo' method")
                if not hasattr(sampler, "sample_ising") or not callable(sampler.sample_ising):
                    raise TypeError("expected sampler to have a 'sample_ising' method")

            # now run the function and return the results
            return f(*new_args, **kw)
        return func
    return decorator



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ["structural_imbalance"]


@binary_quadratic_model_sampler(1)
def structural_imbalance(S, sampler=None, **sampler_args):
    """Returns an approximate set of frustrated edges and a bicoloring.

    A signed social network graph is a graph whose signed edges
    represent friendly/hostile interactions between nodes. A
    signed social network is considered balanced if it can be cleanly
    divided into two factions, where all relations within a faction are
    friendly, and all relations between factions are hostile. The measure
    of imbalance or frustration is the minimum number of edges that
    violate this rule.

    Parameters
    ----------
    S : NetworkX graph
        A social graph on which each edge has a 'sign'
        attribute with a numeric value.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrainted Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    frustrated_edges : dict
        A dictionary of the edges that violate the edge sign. The imbalance
        of the network is the length of frustrated_edges.

    colors: dict
        A bicoloring of the nodes into two factions.

    Raises
    ------
    ValueError
        If any edge does not have a 'sign' attribute.

    Examples
    --------
    >>> import dimod
    >>> sampler = dimod.ExactSolver()
    >>> S = nx.Graph()
    >>> S.add_edge('Alice', 'Bob', sign=1)  # Alice and Bob are friendly
    >>> S.add_edge('Alice', 'Eve', sign=-1)  # Alice and Eve are hostile
    >>> S.add_edge('Bob', 'Eve', sign=-1)  # Bob and Eve are hostile
    >>> frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
    >>> print(frustrated_edges)
    {}
    >>> print(colors)  # doctest: +SKIP
    {'Alice': 0, 'Bob': 0, 'Eve': 1}
    >>> S.add_edge('Ted', 'Bob', sign=1)  # Ted is friendly with all
    >>> S.add_edge('Ted', 'Alice', sign=1)
    >>> S.add_edge('Ted', 'Eve', sign=1)
    >>> frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
    >>> print(frustrated_edges)  # doctest: +SKIP
    {('Ted', 'Eve'): {'sign': 1}}
    >>> print(colors)  # doctest: +SKIP
    {'Bob': 1, 'Ted': 1, 'Alice': 1, 'Eve': 0}

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Ising model on Wikipedia <https://en.wikipedia.org/wiki/Ising_model>`_

    .. [FIA] Facchetti, G., Iacono G., and Altafini C. (2011). Computing
       global structural balance in large-scale signed social networks.
       PNAS, 108, no. 52, 20953-20958

    """
    h, J = structural_imbalance_ising(S)

    # use the sampler to find low energy states
    response = sampler.sample_ising(h, J, **sampler_args)

    # we want the lowest energy sample
    sample = next(iter(response))

    # spins determine the color
    colors = {v: (spin + 1) // 2 for v, spin in sample.items()}

    # frustrated edges are the ones that are violated
    frustrated_edges = {}
    for u, v, data in S.edges(data=True):
        sign = data['sign']

        if sign > 0 and colors[u] != colors[v]:
            frustrated_edges[(u, v)] = data
        elif sign < 0 and colors[u] == colors[v]:
            frustrated_edges[(u, v)] = data
        # else: not frustrated or sign == 0, no relation to violate

    return frustrated_edges, colors


def structural_imbalance_ising(S):
    """Construct the Ising problem to calculate the structural imbalance of a signed social network.

    A signed social network graph is a graph whose signed edges
    represent friendly/hostile interactions between nodes. A
    signed social network is considered balanced if it can be cleanly
    divided into two factions, where all relations within a faction are
    friendly, and all relations between factions are hostile. The measure
    of imbalance or frustration is the minimum number of edges that
    violate this rule.

    Parameters
    ----------
    S : NetworkX graph
        A social graph on which each edge has a 'sign' attribute with a numeric value.

    Returns
    -------
    h : dict
        The linear biases of the Ising problem. Each variable in the Ising problem represent
        a node in the signed social network. The solution that minimized the Ising problem
        will assign each variable a value, either -1 or 1. This bi-coloring defines the factions.

    J : dict
        The quadratic biases of the Ising problem.

    Raises
    ------
    ValueError
        If any edge does not have a 'sign' attribute.

    Examples
    --------
    >>> import dimod
    >>> from dwave_networkx.algorithms.social import structural_imbalance_ising
    ...
    >>> S = nx.Graph()
    >>> S.add_edge('Alice', 'Bob', sign=1)  # Alice and Bob are friendly
    >>> S.add_edge('Alice', 'Eve', sign=-1)  # Alice and Eve are hostile
    >>> S.add_edge('Bob', 'Eve', sign=-1)  # Bob and Eve are hostile
    ...
    >>> h, J = structural_imbalance_ising(S)
    >>> h  # doctest: +SKIP
    {'Alice': 0.0, 'Bob': 0.0, 'Eve': 0.0}
    >>> J  # doctest: +SKIP
    {('Alice', 'Bob'): -1.0, ('Alice', 'Eve'): 1.0, ('Bob', 'Eve'): 1.0}

    """
    h = {v: 0.0 for v in S}
    J = {}
    for u, v, data in S.edges(data=True):
        try:
            J[(u, v)] = -1. * data['sign']
        except KeyError:
            raise ValueError(("graph should be a signed social graph,"
                              "each edge should have a 'sign' attr"))

    return h, J



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Generators for some graphs derived from the D-Wave System.
"""
import re

import networkx as nx

from dwave_networkx.exceptions import DWaveNetworkXException
import warnings

from itertools import product
from .chimera import _chimera_coordinates_cache

__all__ = ['pegasus_graph',
           'pegasus_coordinates',
           'pegasus_sublattice_mappings',
           ]


def pegasus_graph(m, create_using=None, node_list=None, edge_list=None, data=True,
                  offset_lists=None, offsets_index=None, coordinates=False, fabric_only=True,
                  nice_coordinates=False):
    """
    Creates a Pegasus graph with size parameter `m`.

    Parameters
    ----------
    m : int
        Size parameter for the Pegasus lattice.
    create_using : Graph, optional (default None)
        If provided, this graph is cleared of nodes and edges and filled
        with the new graph. Usually used to set the type of the graph.
    node_list : iterable, optional (default None)
        Iterable of nodes in the graph. If None, calculated from `m`.
        Note that this list is used to remove nodes, so any nodes specified
        not in ``range(24 * m * (m-1))`` are not added.
    edge_list : iterable, optional (default None)
        Iterable of edges in the graph. If None, edges are generated as
        described below. The nodes in each edge must be integer-labeled in
        ``range(24 * m * (m-1))``.
    data : bool, optional (default True)
        If True, each node has a pegasus_index attribute. The attribute
        is a 4-tuple Pegasus index as defined below. If the `coordinates` parameter
        is True, a linear_index, which is an integer, is used.
    coordinates : bool, optional (default False)
        If True, node labels are 4-tuple Pegasus indices. Ignored if the
        `nice_coordinates` parameter is True.
    offset_lists : pair of lists, optional (default None)
        Directly controls the offsets. Each list in the pair must have length 12
        and contain even ints.  If `offset_lists` is not None, the `offsets_index`
        parameter must be None.
    offsets_index : int, optional (default None)
        A number between 0 and 7, inclusive, that selects a preconfigured
        set of topological parameters. If both the `offsets_index` and
        `offset_lists` parameters are None, the `offsets_index` parameters is set
        to zero. At least one of these two parameters must be None.
    fabric_only: bool, optional (default True)
        The Pegasus graph, by definition, has some disconnected
        components.  If True, the generator only constructs nodes from the
        largest component. If False, the full disconnected graph is
        constructed. Ignored if the `edge_lists` parameter is not None or
        `nice_coordinates` is True
    nice_coordinates: bool, optional (default False)
        If the `offsets_index` parameter is 0, the graph uses a "nicer"
        coordinate system, more compatible with Chimera addressing.
        These coordinates are 5-tuples taking the form :math:`(t, y, x, u, k)` where
        :math:`0 <= x < M-1`, :math:`0 <= y < M-1`, :math:`0 <= u < 2`,
        :math:`0 <= k < 4`, and :math:`0 <= t < 3`.
        For any given :math:`0 <= t0 < 3`, the subgraph of nodes with :math:`t = t0`
        has the structure of `chimera(M-1, M-1, 4)` with the addition of odd couplers.
        Supercedes both the `fabric_only` and `coordinates` parameters.

    Returns
    -------
    G : NetworkX Graph
        A Pegasus lattice for size parameter `m`.


    The maximum degree of this graph is 15. The number of nodes depends on multiple
    parameters; for example,

        * `pegasus_graph(1)`: zero nodes
        * `pegasus_graph(m, fabric_only=False)`: :math:`24m(m-1)` nodes
        * `pegasus_graph(m, fabric_only=True)`: :math:`24m(m-1)-8(m-1)` nodes
        * `pegasus_graph(m, nice_coordinates=True)`: :math:`24(m-1)^2` nodes

    Counting formulas for edges have a complicated dependency on parameter settings.
    Some example upper bounds are:

        * `pegasus_graph(1, fabric_only=False)`: zero edges
        * `pegasus_graph(m, fabric_only=False)`: :math:`12*(15*(m-1)^2 + m - 3)`
          edges if m > 1

    Note that the formulas above are valid for default offset parameters.

    A Pegasus lattice is a graph minor of a lattice similar
    to Chimera, where unit tiles are completely connected. In its most general
    definition, prelattice :math:`Q(N0,N1)` contains nodes of the form

        * vertical nodes: :math:`(i, j, 0, k)` with :math:`0 <= k < 2`
        * horizontal nodes: :math:`(i, j, 1, k)` with :math:`0 <= k < 2`

    for :math:`0 <= i <= N0` and :math:`0 <= j < N1`, and edges of the form

        * external: :math:`(i, j, u, k)` ~ :math:`(i+u, j+1-u, u, k)`
        * internal: :math:`(i, j, 0, k)` ~ :math:`(i, j, 1, h)`
        * odd: :math:`(i, j, u, 0)` ~ :math:`(i, j, u, 1)`

    Given two lists of offsets, :math:`S0` and :math:`S1`, of length
    :math:`L0` and :math:`L1`, where both lengths and values must be divisible by
    2, the minor---a Pegasus lattice---is constructed by contracting the complete
    intervals of external edges::

        I(0, w, k, z) = [(L1*w + k, L0*z + S0[k] + r, 0, k % 2) for 0 <= r < L0]
        I(1, w, k, z) = [(L1*z + S1[k] + r, L0*w + k, 1, k % 2) for 0 <= r < L1]

    and deleting the prelattice nodes of any interval not fully contained in
    :math:`Q(N0, N1)`.

    This generator, 'pegasus_graph()', is specialized for the minor constructed by
    prelattice and offset parameters :math:`L0 = L1 = 12` and :math:`N0 = N1 = 12m`.

    The *Pegasus index* of a node in a Pegasus lattice, :math:`(u, w, k, z)`, can be
    interpreted as:

        * :math:`u`: qubit orientation (0 = vertical, 1 = horizontal)
        * :math:`w`: orthogonal major offset
        * :math:`k`: orthogonal minor offset
        * :math:`z`: parallel offset

    Edges in the minor have the form

        * external: :math:`(u, w, k, z)` ~ :math:`(u, w, k, z+1)`
        * internal: :math:`(0, w0, k0, z0)` ~ :math:`(1, w1, k1, z1)`
        * odd: :math:`(u, w, 2k, z)` ~ :math:`(u, w, 2k+1, z)`

    where internal edges only exist when

        1. w1 = z0 + (1 if k1 < S0[k0] else 0)
        2. z1 = w0 - (1 if k0 < S1[k1] else 0)

    Linear indices are computed from Pegasus indices by the formula::

        q = ((u * m + w) * 12 + k) * (m - 1) + z


    Examples
    ========
    >>> G = dnx.pegasus_graph(2, nice_coordinates=True)
    >>> G.nodes(data=True)[(0, 0, 0, 0, 0)]    # doctest: +SKIP
    {'linear_index': 4, 'pegasus_index': (0, 0, 4, 0)}


    """
    if offset_lists is None:
        offsets_descriptor = offsets_index = offsets_index or 0
        offset_lists = [
            [(2, 2, 2, 2, 10, 10, 10, 10, 6, 6, 6, 6,), (6, 6, 6, 6, 2, 2, 2, 2, 10, 10, 10, 10,)],
            [(2, 2, 2, 2, 10, 10, 10, 10, 6, 6, 6, 6,), (2, 2, 2, 2, 10, 10, 10, 10, 6, 6, 6, 6,)],
            [(2, 2, 2, 2, 10, 10, 10, 10, 6, 6, 6, 6,), (10, 10, 10, 10, 6, 6, 6, 6, 2, 2, 2, 2,)],
            [(10, 10, 10, 10, 6, 6, 6, 6, 2, 2, 2, 2,), (10, 10, 10, 10, 6, 6, 6, 6, 2, 2, 2, 2,)],
            [(10, 10, 10, 10, 6, 6, 6, 6, 2, 2, 2, 2,), (2, 2, 2, 2, 6, 6, 6, 6, 10, 10, 10, 10,)],
            [(6, 6, 2, 2, 2, 2, 10, 10, 10, 10, 6, 6,), (6, 6, 2, 2, 2, 2, 10, 10, 10, 10, 6, 6,)],
            [(6, 6, 2, 2, 2, 2, 10, 10, 10, 10, 6, 6,), (6, 6, 10, 10, 10, 10, 2, 2, 2, 2, 6, 6,)],
            [(6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,), (6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,)],
        ][offsets_index]
    elif offsets_index is not None:
        raise DWaveNetworkXException("provide at most one of offsets_index and offset_lists")
    else:
        for ori in 0, 1:
            for x, y in zip(offset_lists[ori][::2], offset_lists[ori][1::2]):
                if x != y:
                    warnings.warn("The offets list you've provided is possibly non-physical.  Odd-coupled qubits should have the same value.")
        offsets_descriptor = offset_lists

    G = nx.empty_graph(0, create_using)

    G.name = "pegasus_graph(%s, %s)" % (m, offsets_descriptor)

    m1 = m - 1
    if nice_coordinates:
        if offsets_index != 0:
            raise NotImplementedError("nice coordinate system is only implemented for offsets_index 0")
        labels = 'nice'
        pegasus_to_nice = pegasus_coordinates.pegasus_to_nice
        nice_to_pegasus = pegasus_coordinates.nice_to_pegasus
        label = lambda *q: pegasus_to_nice(q)
    elif coordinates:
        label = lambda *q: q
        labels = 'coordinate'
    else:
        labels = 'int'
        def label(u, w, k, z):
            return u * 12 * m * m1 + w * 12 * m1 + k * m1 + z

    construction = (("family", "pegasus"), ("rows", m), ("columns", m),
                    ("tile", 12), ("vertical_offsets", offset_lists[0]),
                    ("horizontal_offsets", offset_lists[1]), ("data", data),
                    ("labels", labels))

    G.graph.update(construction)

    max_size = m * (m - 1) * 24  # max number of nodes G can have

    if edge_list is None:
        if nice_coordinates:
            fabric_start = 4,8
            fabric_end = 8, 4
        elif fabric_only:
            fabric_start = min(s for s in offset_lists[1]), min(s for s in offset_lists[0])
            fabric_end = 12 - max(s for s in offset_lists[1]), 12 - max(s for s in offset_lists[0])
        else:
            fabric_end = fabric_start = 0, 0

        G.add_edges_from((label(u, w, k, z), label(u, w, k, z + 1))
                         for u in (0, 1)
                         for w in range(m)
                         for k in range(fabric_start[u] if w == 0 else 0, 12 - (fabric_end[u] if w == m1 else 0))
                         for z in range(m1 - 1))

        G.add_edges_from((label(u, w, k, z), label(u, w, k + 1, z))
                         for u in (0, 1)
                         for w in range(m)
                         for k in range(fabric_start[u] if w == 0 else 0, 12 - (fabric_end[u] if w == m1 else 0), 2)
                         for z in range(m1))

        off0, off1 = offset_lists
        def qfilter(u, w, k, z):
            if w == 0: return k >= fabric_start[u]
            if w == m1: return k < 12-fabric_end[u]
            return True
        def efilter(e): return qfilter(*e[0]) and qfilter(*e[1])

        internal_couplers = (((0, w, k, z), (1, z + (kk < off0[k]), kk, w - (k < off1[kk])))
                         for w in range(m)
                         for kk in range(12)
                         for k in range(0 if w else off1[kk], 12 if w < m1 else off1[kk])
                         for z in range(m1))
        G.add_edges_from((label(*e[0]), label(*e[1])) for e in internal_couplers if efilter(e))

    else:
        G.add_edges_from(edge_list)

    if node_list is not None:
        nodes = set(node_list)
        G.remove_nodes_from(set(G) - nodes)
        G.add_nodes_from(nodes)  # for singleton nodes

    if data:
        v = 0
        if nice_coordinates:
            def fill_data():
                q = (u, w, k, z)
                d = get_node_data(pegasus_to_nice(q))
                if d is not None:
                    d['linear_index'] = v
                    d['pegasus_index'] = q
        elif coordinates:
            def fill_data():
                d = get_node_data((u, w, k, z))
                if d is not None:
                    d['linear_index'] = v
        else:
            def fill_data():
                d = get_node_data(v)
                if d is not None:
                    d['pegasus_index'] = (u, w, k, z)

        get_node_data = G.nodes.get
        for u in range(2):
            for w in range(m):
                for k in range(12):
                    for z in range(m1):
                        fill_data()
                        v += 1

    return G


def get_tuple_fragmentation_fn(pegasus_graph):
    """
    Returns a fragmentation function that is specific to pegasus_graph. This fragmentation function,
    fragment_tuple(..), takes in a list of Pegasus qubit coordinates and returns their corresponding
    K2,2 Chimera fragment coordinates.

    Details on the returned function, fragment_tuple(list_of_pegasus_coordinates):
        Each Pegasus qubit is split into six fragments. If edges are drawn between adjacent
        fragments and drawn between fragments that are connected by an existing Pegasus coupler, we
        can see that a K2,2 Chimera graph is formed.

        The K2,2 Chimera graph uses a coordinate system with an origin at the upper left corner of
        the graph.
            y: number of vertical fragments from the top-most row
            x: number of horizontal fragments from the left-most column
            u: 1 if it belongs to a horizontal qubit, 0 otherwise
            r: fragment index on the K2,2 shore

    Parameters
    ----------
    pegasus_graph: networkx.graph
        A pegasus graph

    Returns
    -------
    fragment_tuple(pegasus_coordinates): a function
        A function that accepts a list of pegasus coordinates and returns a list of their
        corresponding K2,2 Chimera coordinates.
    """
    horizontal_offsets = pegasus_graph.graph['horizontal_offsets']
    vertical_offsets = pegasus_graph.graph['vertical_offsets']

    # Note: we are returning a fragmentation function rather than fragmenting the pegasus
    # coordinates ourselves because:
    #   (1) We don't want the user to have to deal with Pegasus horizontal/vertical offsets directly.
    #       (i.e. Don't want fragment_tuple(pegasus_coord, vertical_offset, horizontal_offset))
    #   (2) We don't want the user to have to pass entire Pegasus graph each time they want to
    #       fragment some pegasus coordinates.
    #       (i.e. Don't want fragment_tuple(pegasus_coord, pegasus_graph))
    def fragment_tuple(pegasus_coords):
        fragments = []
        for u, w, k, z in pegasus_coords:
            # Determine offset
            offset = horizontal_offsets if u else vertical_offsets
            offset = offset[k]

            # Find the base (i.e. zeroth) Chimera fragment of this pegasus coordinate
            fz0 = (z*12 + offset) // 2 #first fragment's z-coordinate
            fw = (w*12 + k) // 2 #fragment w-coordinate
            fk = k&1 #fragment k-index
            base = [fw, 0, u, fk] if u else [0, fw, u, fk]

            # Generate the six fragments associated with this pegasus coordinate
            for fz in range(fz0, fz0 + 6):
                base[u] = fz
                fragments.append(tuple(base))

        return fragments

    return fragment_tuple


def get_tuple_defragmentation_fn(pegasus_graph):
    """
    Returns a de-fragmentation function that is specific to pegasus_graph. The returned
    de-fragmentation function, defragment_tuple(..), takes in a list of K2,2 Chimera coordinates and
    returns the corresponding list of unique pegasus coordinates.

    Details on the returned function, defragment_tuple(list_of_chimera_fragment_coordinates):
        Each Pegasus qubit is split into six fragments. If edges are drawn between adjacent
        fragments and drawn between fragments that are connected by an existing Pegasus coupler, we
        can see that a K2,2 Chimera graph is formed.

        The K2,2 Chimera graph uses a coordinate system with an origin at the upper left corner of
        the graph.
            y: number of vertical fragments from the top-most row
            x: number of horizontal fragments from the left-most column
            u: 1 if it belongs to a horizontal qubit, 0 otherwise
            r: fragment index on the K2,2 shore

        The defragment_tuple(..) takes in the list of Chimera fragments and returns a list of their
        corresponding Pegasus qubit coordinates. Note that the returned list has a unique set of
        Pegasus coordinates.

    Parameters
    ----------
    pegasus_graph: networkx.graph
        A Pegasus graph

    Returns
    -------
    defragment_tuple(chimera_coordinates): a function
        A function that accepts a list of chimera coordinates and returns a set of their
        corresponding Pegasus coordinates.
    """
    horizontal_offsets = pegasus_graph.graph['horizontal_offsets']
    vertical_offsets = pegasus_graph.graph['vertical_offsets']

    # Note: we are returning a defragmentation function rather than defragmenting the chimera
    # fragments ourselves because:
    #   (1) We don't want the user to have to deal with Pegasus horizontal/vertical offsets directly.
    #       (i.e. Don't want defragment_tuple(chimera_coord, vertical_offset, horizontal_offset))
    #   (2) We don't want the user to have to pass entire Pegasus graph each time they want to
    #       defragment some chimera coordinates.
    #       (i.e. Don't want defragment_tuple(chimera_coord, pegasus_graph))
    def defragment_tuple(chimera_coords):
        pegasus_coords = []
        for y, x, u, r in chimera_coords:
            # Set up shifts and offsets
            shifts = [x, y]
            offsets = horizontal_offsets if u else vertical_offsets

            # Determine number of tiles and track number
            w, k = divmod(2 * shifts[u] + r, 12)

            # Determine qubit index on track
            z = (shifts[1-u] * 2 - offsets[k]) // 12

            pegasus_coords.append((u, w, k, z))

        # Several chimera coordinates may map to the same pegasus coordinate, hence, apply set(..)
        return list(set(pegasus_coords))

    return defragment_tuple


def fragmented_edges(pegasus_graph):
    """
    Generator for the edges contained in a Chimera graph obtained by splitting each Pegasus node into
    six Chimera nodes.  If the Pegasus graph has size parameter m, then the derived graph will be a
    subgraph of Chimera(6m, 6m, 2) -- that is, a Chimera graph with K2,2 unit tiles.

        The K2,2 Chimera graph uses a coordinate system with an origin at the upper left corner of
        the graph.
            y: number of vertical fragments from the top-most row
            x: number of horizontal fragments from the left-most column
            u: 1 if it belongs to a horizontal qubit, 0 otherwise
            r: fragment index on the K2,2 shore

    Parameters
    ----------
    pegasus_graph: networkx.graph
        A pegasus graph

    Returns
    -------
    (coord0, coord1), ... : an iterator of tuples
        Yields the edges contained in the Chimera graph derived from the "fragmentation" construction
    """
    offsetlist = [pegasus_graph.graph['vertical_offsets'], pegasus_graph.graph['horizontal_offsets']]
    offsets = {(u, k): offsetlist[u][k] for u in (0, 1) for k in range(12)}

    if pegasus_graph.graph['labels'] == 'nice':
        coords = pegasus_coordinates.nice_to_pegasus
    elif pegasus_graph.graph['labels'] == 'int':
        coords = pegasus_coordinates(pegasus_graph.graph['rows']).linear_to_pegasus
    else:
        coords = lambda z: z

    #first, we generate the edges internal to the fragments corresponding to a node
    for q in pegasus_graph.nodes():
        u, w, k, z = coords(q)
        #copied from get_tuple_fragmentation_fn and slightly optimized
        offset = offsets[u, k]
        fz0 = (z*12 + offset) // 2 #first fragment z-coordinate
        fw = (w*12 + k) // 2 #fragment w-coordinate
        base = [fw, fz0, u, k&1] if u else [fz0, fw, u, k&1]
        prev = tuple(base)
        for fz in range(fz0+1, fz0+6):
            base[u] = fz
            curr = tuple(base)
            yield (prev, curr)
            prev = curr

    #now for the thinky part: for each Pegasus edge, generate the corresponding Chimera edge
    #we skip the "odd-coupler" edges because they don't exist in Chimera
    for q0, q1 in pegasus_graph.edges():
        u0, w0, k0, z0 = coords(q0)
        u1, w1, k1, z1 = coords(q1)
        if u0 == u1:
            if k0 == k1:
                #this is an external edge -- we could probably do some hijinks to fold this in
                #with the nodes loop, but cost/benefit doesn't support it right now
                offset = offsets[u0, k0]
                fz = (min(z0, z1)*12 + offset) // 2 #first fragment z-coordinate in the pair
                fw = (w0*12 + k0) // 2 #fragment w-coordinate for both qubits
                fk = k0&1 #fragment k-index
                if u0:
                    yield ((fw, fz+5, u0, fk), (fw, fz+6, u0, fk))
                else:
                    yield ((fz+5, fw, u0, fk), (fz+6, fw, u0, fk))

            #else: this is an odd edge; yield nothing
        else:
            #this may look a little magical -- we're looking for an edge of the form
            # (fy, fx, u0, fk0), (fy, fx, u1, fk1)
            #where (fy, fx) are the first two coordinates of both the fragments of (u0, w0, k0,z0),
            # (fy, fx) in [(fz0+0, fw0), (fz0+1, fw0), ..., (fz0+5, fw0)]
            #and also the the fragments of (u1, w1, k1, z1):
            # (fy, fx) in [(fw1, fz1+0), (fw1, fz1+1), ..., (fw1, fz1+5)]
            #(see get_tuple_fragmentation_fn to see the fragment generator)
            #with the assumption that an intersection exists: it can only be located at (fw0, fw1)
            #since those coordinates are constant in the respective intervals.  Thus, we get to
            #skip looking up the offsets.  Magic?  No, Math!
            fw0 = (w0*12 + k0) // 2
            fw1 = (w1*12 + k1) // 2
            if u0:
                yield ((fw0, fw1, u0, k0&1), (fw0, fw1, u1, k1&1))
            else:
                yield ((fw1, fw0, u0, k0&1), (fw1, fw0, u1, k1&1))


# Developer note: we could implement a function that creates the iter_*_to_* and
# iter_*_to_*_pairs methods just-in-time, but there are a small enough number
# that for now it makes sense to do them by hand.
class pegasus_coordinates(object):
    """Provides coordinate converters for the Pegasus indexing schemes.

    Parameters
    ----------
    m : int
        Size parameter for the Pegasus lattice.

    See also
    --------
    :func:`.pegasus_graph` : Describes the various coordinate conventions.

    """
    def __init__(self, m):

        self.args = m, m - 1

    def pegasus_to_linear(self, q):
        """Convert a 4-term Pegasus coordinate into a linear index.

        Parameters
        ----------
        q : 4-tuple
            Pegasus indices.

        Examples
        --------
        >>> dnx.pegasus_coordinates(2).pegasus_to_linear((0, 0, 4, 0))
        4
        """
        u, w, k, z = q
        m, m1 = self.args
        return ((m * u + w) * 12 + k) * m1 + z

    def linear_to_pegasus(self, r):
        """Convert a linear index into a 4-term Pegasus coordinate.

        Parameters
        ----------
        r : int
            Linear index.

        Examples
        --------
        >>> dnx.pegasus_coordinates(2).linear_to_pegasus(4)
        (0, 0, 4, 0)

        """
        m, m1 = self.args
        r, z = divmod(r, m1)
        r, k = divmod(r, 12)
        u, w = divmod(r, m)
        return u, w, k, z

    @staticmethod
    def nice_to_pegasus(n):
        """Convert a 5-term nice coordinate into a 4-term Pegasus coordinate.

        Parameters
        ----------
        n : 5-tuple
            Nice coordinate.

        Examples
        --------
        >>> dnx.pegasus_coordinates.nice_to_pegasus((0, 0, 0, 0, 0))
        (0, 0, 4, 0)

        Note that this method does not depend on the size of the Pegasus
        lattice.
        """
        t, y, x, u, k = n

        if t == 0:
            return (u, y+1 if u else x, 4+k if u else 4+k, x if u else y)
        elif t == 1:
            return (u, y+1 if u else x, k if u else 8+k, x if u else y)
        elif t == 2:
            return (u, y if u else x + 1, 8+k if u else k, x if u else y)

        # can happen when t is a float for instance
        raise ValueError("invalid Nice coordinate: {}".format(n))

    @staticmethod
    def pegasus_to_nice(p):
        """Convert a 4-term Pegasus coordinate to a 5-term nice coordinate.

        Parameters
        ----------
        p : 4-tuple
            Pegasus coordinate.

        Examples
        --------
        >>> dnx.pegasus_coordinates.pegasus_to_nice((0, 0, 4, 0))
        (0, 0, 0, 0, 0)

        Note that this method does not depend on the size of the Pegasus
        lattice.
        """
        u, w, k, z = p

        t = (2-u-(2*u-1)*(k//4)) % 3

        if t == 0:
            return (0, w-1 if u else z, z if u else w, u, k-4 if u else k-4)
        elif t == 1:
            return (1, w-1 if u else z, z if u else w, u, k if u else k-8)
        elif t == 2:
            return (2, w if u else z, z if u else w-1, u, k-8 if u else k)

        # can happen when given floats for instance
        raise ValueError('invalid Pegasus coordinates')

    def linear_to_nice(self, r):
        """Convert a linear index into a 5-term nice coordinate.

        Parameters
        ----------
        r : int
            Linear index.

        Examples
        --------
        >>> dnx.pegasus_coordinates(2).linear_to_nice(4)
        (0, 0, 0, 0, 0)
        """
        return self.pegasus_to_nice(self.linear_to_pegasus(r))

    def nice_to_linear(self, n):
        """Convert a 5-term nice coordinate into a linear index.

        Parameters
        ----------
        n : 5-tuple
            Nice coordinate.

        Examples
        --------
        >>> dnx.pegasus_coordinates(2).nice_to_linear((0, 0, 0, 0, 0))
        4
        """
        return self.pegasus_to_linear(self.nice_to_pegasus(n))

    def iter_pegasus_to_linear(self, qlist):
        """Return an iterator converting a sequence of 4-term Pegasus
        coordinates to linear indices.
        """
        m, m1 = self.args
        for (u, w, k, z) in qlist:
            yield ((m * u + w) * 12 + k) * m1 + z

    def iter_linear_to_pegasus(self, rlist):
        """Return an iterator converting a sequence of linear indices to 4-term
        Pegasus coordinates.
        """
        m, m1 = self.args
        for r in rlist:
            r, z = divmod(r, m1)
            r, k = divmod(r, 12)
            u, w = divmod(r, m)
            yield u, w, k, z

    @classmethod
    def iter_nice_to_pegasus(cls, nlist):
        """Return an iterator converting a sequence of 5-term nice coordinates
        to 4-term Pegasus coordinates.

        Note that this method does not depend on the size of the Pegasus
        lattice.
        """
        for n in nlist:
            yield cls.nice_to_pegasus(n)

    @classmethod
    def iter_pegasus_to_nice(cls, plist):
        """Return an iterator converting a sequence of 4-term Pegasus
        coordinates to 5-term nice coordinates.

        Note that this method does not depend on the size of the Pegasus
        lattice.
        """
        for p in plist:
            yield cls.pegasus_to_nice(p)

    def iter_linear_to_nice(self, rlist):
        """Return an iterator converting a sequence of linear indices to 5-term
        nice coordinates.
        """
        for r in rlist:
            yield self.linear_to_nice(r)

    def iter_nice_to_linear(self, nlist):
        """Return an iterator converting a sequence of 5-term nice coordinates
        to linear indices.
        """
        for n in nlist:
            yield self.nice_to_linear(n)

    @staticmethod
    def _pair_repack(f, plist):
        """Flattens a sequence of pairs to pass through `f`, and then
        re-pairs the result.
        """
        ulist = f(u for p in plist for u in p)
        for u in ulist:
            v = next(ulist)
            yield u, v

    def iter_pegasus_to_linear_pairs(self, plist):
        """Return an iterator converting a sequence of pairs of 4-term Pegasus
        coordinates to pairs of linear indices.
        """
        return self._pair_repack(self.iter_pegasus_to_linear, plist)

    def iter_linear_to_pegasus_pairs(self, plist):
        """Return an iterator converting a sequence of pairs of linear indices
        to pairs of 4-term Pegasus coordinates.
        """
        return self._pair_repack(self.iter_linear_to_pegasus, plist)

    @classmethod
    def iter_nice_to_pegasus_pairs(cls, nlist):
        """Return an iterator converting a sequence of pairs of 5-term nice
        coordinates to pairs of 4-term Pegasus coordinates.

        Note that this method does not depend on the size of the Pegasus
        lattice.
        """
        return cls._pair_repack(cls.iter_nice_to_pegasus, nlist)

    @classmethod
    def iter_pegasus_to_nice_pairs(cls, plist):
        """Return an iterator converting a sequence of pairs of 4-term Pegasus
        coordinates to pairs of 5-term nice coordinates.

        Note that this method does not depend on the size of the Pegasus
        lattice.
        """
        return cls._pair_repack(cls.iter_pegasus_to_nice, plist)

    def iter_linear_to_nice_pairs(self, rlist):
        """Return an iterator converting a sequence of pairs of linear indices
        to pairs of 5-term nice coordinates.
        """
        return self._pair_repack(self.iter_linear_to_nice, rlist)

    def iter_nice_to_linear_pairs(self, nlist):
        """Return an iterator converting a sequence of pairs of 5-term nice
        coordinates to pairs of linear indices.
        """
        return self._pair_repack(self.iter_nice_to_linear, nlist)

    def graph_to_linear(self, g):
        """Return a copy of the graph g relabeled to have linear indices"""
        labels = g.graph.get('labels')
        if labels == 'int':
            return g.copy()
        elif labels == 'coordinate':
            nodes = self.iter_pegasus_to_linear(g)
            edges = self.iter_pegasus_to_linear_pairs(g.edges)
        elif labels == 'nice':
            nodes = self.iter_nice_to_linear(g)
            edges = self.iter_nice_to_linear_pairs(g.edges)
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.pegasus_graph."
            )

        return pegasus_graph(
            g.graph['rows'],
            node_list=nodes,
            edge_list=edges,
            data=g.graph['data'],
            offset_lists=(
                g.graph["vertical_offsets"],
                g.graph["horizontal_offsets"],
            ),
        )

    def graph_to_pegasus(self, g):
        """Return a copy of the graph g relabeled to have pegasus coordinates"""
        labels = g.graph.get('labels')
        if labels == 'int':
            nodes = self.iter_linear_to_pegasus(g)
            edges = self.iter_linear_to_pegasus_pairs(g.edges)
        elif labels == 'coordinate':
            return g.copy()
        elif labels == 'nice':
            nodes = self.iter_nice_to_pegasus(g)
            edges = self.iter_nice_to_pegasus_pairs(g.edges)
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.pegasus_graph."
            )

        return pegasus_graph(
            g.graph['rows'],
            node_list=nodes,
            edge_list=edges,
            data=g.graph['data'],
            coordinates=True,
            offset_lists=(
                g.graph["vertical_offsets"],
                g.graph["horizontal_offsets"],
            ),
        )

    def graph_to_nice(self, g):
        """Return a copy of the graph p relabeled to have nice coordinates"""
        labels = g.graph.get('labels')
        if labels == 'int':
            nodes = self.iter_linear_to_nice(g)
            edges = self.iter_linear_to_nice_pairs(g.edges)
        elif labels == 'coordinate':
            nodes = self.iter_pegasus_to_nice(g)
            edges = self.iter_pegasus_to_nice_pairs(g.edges)
        elif labels == 'nice':
            return g.copy()
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.pegasus_graph."
            )

        return pegasus_graph(
            g.graph['rows'],
            node_list=nodes,
            edge_list=edges,
            data=g.graph['data'],
            nice_coordinates=True,
            offsets_index = 0,
        )

    def int(self, q):
        """Deprecated alias of `pegasus_to_linear`."""
        msg = ('pegasus_coordinates.int is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'pegasus_coordinates.pegasus_to_linear instead')
        warnings.warn(msg, DeprecationWarning)
        return self.pegasus_to_linear(q)

    def tuple(self, r):
        """Deprecated alias for `linear_to_pegasus`."""
        msg = ('pegasus_coordinates.tuple is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'pegasus_coordinates.linear_to_pegasus instead')
        warnings.warn(msg, DeprecationWarning)
        return self.linear_to_pegasus(r)

    def ints(self, qlist):
        """Deprecated alias for `iter_pegasus_to_linear`."""
        msg = ('pegasus_coordinates.ints is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'pegasus_coordinates.iter_pegasus_to_linear instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_pegasus_to_linear(qlist)

    def tuples(self, rlist):
        """Deprecated alias for `iter_linear_to_pegasus`."""
        msg = ('pegasus_coordinates.tuples is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'pegasus_coordinates.iter_linear_to_pegasus instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_linear_to_pegasus(rlist)

    def int_pairs(self, plist):
        """Deprecated alias for `iter_pegasus_to_linear_pairs`."""
        msg = ('pegasus_coordinates.int_pairs is deprecated and will be removed'
               ' in dwave-networkx 0.9.0, please use '
               'pegasus_coordinates.iter_pegasus_to_linear_pairs instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_pegasus_to_linear_pairs(plist)

    def tuple_pairs(self, plist):
        """Deprecated alias for `iter_linear_to_pegasus_pairs`."""
        msg = ('pegasus_coordinates.tuple_pairs is deprecated and will be removed'
               ' in dwave-networkx 0.9.0, please use '
               'pegasus_coordinates.iter_linear_to_pegasus_pairs instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_linear_to_pegasus_pairs(plist)


# maintained for backwards compatibility
def get_pegasus_to_nice_fn(*args, **kwargs):
    """
    Returns a coordinate translation function from the 4-term pegasus_index
    coordinates to the 5-term "nice" coordinates.

    Details on the returned function, pegasus_to_nice(u,w,k,z)
        Inputs are 4-tuples of ints, return is a 5-tuple of ints.  See
        pegasus_graph for description of the pegasus_index and "nice"
        coordinate systems.

    Returns
    -------
    pegasus_to_nice_fn(chimera_coordinates): a function
        A function that accepts augmented chimera coordinates and returns corresponding
        Pegasus coordinates.
    """
    if args or kwargs:
        msg = "get_pegasus_to_nice_fn does not need / use parameters anymore"
        warnings.warn(msg, DeprecationWarning)

    msg = ('get_pegasus_to_nice_fn is deprecated and will be removed in '
           'dwave-networkx 0.9.0, please use '
           'pegasus_coordinates.pegasus_to_nice instead')
    warnings.warn(msg, DeprecationWarning)

    return lambda *args: pegasus_coordinates.pegasus_to_nice(args)


# maintained for backwards compatibility
def get_nice_to_pegasus_fn(*args, **kwargs):
    """
    Returns a coordinate translation function from the 5-term "nice"
    coordinates to the 4-term pegasus_index coordinates.

    Details on the returned function, nice_to_pegasus(t, y, x, u, k)
        Inputs are 5-tuples of ints, return is a 4-tuple of ints.  See
        pegasus_graph for description of the pegasus_index and "nice"
        coordinate systems.

    Returns
    -------
    nice_to_pegasus_fn(pegasus_coordinates): a function
        A function that accepts Pegasus coordinates and returns the corresponding
        augmented chimera coordinates
    """
    if args or kwargs:
        msg = "get_pegasus_to_nice_fn does not need / use parameters anymore"
        warnings.warn(msg, DeprecationWarning)

    msg = ('get_nice_to_pegasus_fn is deprecated and will be removed in '
           'dwave-networkx 0.9.0, please use '
           'pegasus_coordinates.nice_to_pegasus instead')
    warnings.warn(msg, DeprecationWarning)

    return lambda *args: pegasus_coordinates.nice_to_pegasus(args)


def _chimera_pegasus_sublattice_mapping(source_to_chimera, nice_to_target, offset):
    """Constructs a mapping from a Chimera graph to a Pegasus graph, via an offset.
    This function is used by pegasus_sublattice_mappings, and serves to 
    construct a closure that is stable under iteration therein.

    Parameters
    ----------
        source_to_chimera : function
            A function mapping a source node to a chimera coordinate
        nice_to_target: function
            A function mapping a pegasus nice-coordinate to a target node
        offset : tuple (int, int, int)
            A triplet of ints representing the t-, y- and x-offset of the
            sublattice.

    Returns
    -------
        mapping : function
            The function implementing the mapping from the source Chimera
            graph to the target Pegasus graph.  We store ``offset`` in the
            attribute ``mapping.offset`` for later reconstruction.
        
    """
    t_offset, y_offset, x_offset = offset

    def mapping(q):
        y, x, u, k = source_to_chimera(q)
        return nice_to_target((t_offset, y + y_offset, x + x_offset, u, k))

    #store the offset in the mapping, so the user can reconstruct it
    mapping.offset = offset

    return mapping


class __pegasus_coordinates_cache_dict(dict):
    """An internal-use cached factory for `pegasus_coordinates` objects"""
    def __missing__(self, key):
        self[key] = val = pegasus_coordinates(key)
        return val

_pegasus_coordinates_cache = __pegasus_coordinates_cache_dict()

# a set of manually-computed values to speed up the generation of sublattice
# mappings, used in _pegasus_pegasus_sublattice_mapping
_sublattice_delta_list = (
    ((0, 0, 0), (1, 0, 0), (2, 0, 0)),
    ((1, 1, 0), (2, 1, 0), (0, 0, 1)),
    ((2, 1, 0), (0, 0, 1), (1, 0, 1)),
)

def _pegasus_pegasus_sublattice_mapping(source_to_nice, nice_to_target, offset):
    """Constructs a mapping from a Pegasus graph to a Pegasus graph, via an offset.
    This function is used by pegasus_sublattice_mappings, and serves to 
    construct a closure that is stable under iteration therein.

    Parameters
    ----------
        source_to_nice : function
            A function mapping a source node to a pegasus nice-coordinate
        nice_to_target: function
            A function mapping a pegasus nice-coordinate to a target node.
        offset : tuple (int, int, int)
            A triplet of ints representing the t-, y- and x-offset of the
            sublattice.

    Returns
    -------
        mapping : function
            The function implementing the mapping from the source Pegasus
            graph to the target Pegasus graph.  We store ``offset`` in the
            attribute ``mapping.offset`` for later reconstruction.
        
    """
    t_offset, y_offset, x_offset = offset
    delta = _sublattice_delta_list[t_offset]
    def mapping(q):
        T, Y, X, u, k = source_to_nice(q)
        t, dy, dx = delta[T]
        return nice_to_target((t, Y + dy + y_offset, X + dx + x_offset, u, k))

    #store the offset in the mapping, so the user can reconstruct it
    mapping.offset = offset

    return mapping


def pegasus_sublattice_mappings(source, target, offset_list=None):
    """Yields mappings from a Chimera or Pegasus graph into a Pegasus graph.
    
    A sublattice mapping is a function from nodes of a ``pegasus_graph(m_s)`` or
    ``chimera_graph(m_c, n_c, 4)`` to nodes of a ``pegasus_graph(m_t)`` with
    ``m_s <= m_t`` or ``m_c <= m_t - 1`` and ``n_c <= m_t - 1``.  This is used
    to identify subgraphs of the target Pegasus graphs which are isomorphic to
    the source graph.  However, if the target graph is not of perfect yield,
    these functions do not generally produce isomorphisms (for example, if a 
    node is missing in the target graph, it may still appear in the image of the
    source graph).
    
    Note that we require the tile parameter of Chimera graphs to be 4, and the
    mappings produced are not exhaustive.  The mappings take the form
    
        ``(y, x, u, k) -> (t_offset, y+y_offset, x+x_offset, u, k)``
        
    when the source is a Chimera graph, or
    
        ``(t, y, x, u, k) -> ((t + t_offset)%3, y+y_offset, x+x_offset, u, k)``
    
    when the source is a Pegasus graph; preserving the orientation and tile
    index of nodes.  We use the notation of Chimera coordinates and Pegasus nice 
    coordinates above, but the mapping produced will respect the labelings of
    the source and target graph.  Note, the notation above for Pegasus->Pegasus
    mappings is only suggestive. See _pegasus_pegasus_sublattice_mapping for a
    precise formula.
    
    Academic note: the full group of isomorphisms of a Chimera graph includes 
    mappings which permute tile indices on a per-row and per-column basis, in
    addition to reflections and rotations of the grid of unit cells where 
    rotations by 90 and 270 degrees induce a change in orientation.  The
    isomorphisms of Pegasus graphs permit the swapping across rows and columns
    of odd couplers, as well as a reflection about the main antidiagonal which
    induces a change in orientation.  The full set of sublattice mappings would
    take those isomorphisms into account; we do not undertake that complexity
    here.

    Parameters
    ----------
        source : NetworkX Graph
            The Chimera or Pegasus graph that nodes are input from
        target : NetworkX Graph
            The Pegasus graph that nodes are output to
        offset_list : iterable (tuple), optional (default None)
            An iterable of offsets.  This can be used to reconstruct a set of
            mappings, as the offset used to generate a single mapping is stored
            in the ``offset`` attribute of that mapping.
            
    Yields
    ------
        mapping : function
            A function from nodes of the source graph, to nodes of the target
            graph.  The offset used to generate this mapping is stored in
            ``mapping.offset`` -- these can be collected and passed into 
            ``offset_list`` in a later session.
        
    """
    if target.graph.get('family') != 'pegasus':
        raise ValueError("source graphs must a Pegasus graph constructed by dwave_networkx.pegasus_graph")

    m_t = target.graph['rows']
    labels_t = target.graph['labels']
    if labels_t == 'int':
        nice_to_target = _pegasus_coordinates_cache[m_t].nice_to_linear
    elif labels_t == 'coordinate':
        nice_to_target = _pegasus_coordinates_cache[m_t].nice_to_pegasus
    elif labels_t == 'nice':
        def nice_to_target(q):
            return q
    else:
        raise ValueError(f"Pegasus node labeling {labels_t} not recognized")

    labels_s = source.graph['labels']    
    if source.graph.get('family') == 'chimera':
        if source.graph['tile'] != 4:
            raise ValueError("Cannot construct sublattice mappings from Chimera to Pegasus unless the Chimera tile parameter is 4")

        m_s = source.graph['rows']
        n_s = source.graph['columns']
        if offset_list is None:
            offset_list = product([0, 1, 2], range(m_t - m_s), range(m_t - n_s))
        if labels_s == 'coordinate':
            def source_to_inner(q):
                return q
        elif labels_s == 'int':
            source_to_inner = _chimera_coordinates_cache[m_s, n_s, 4].linear_to_chimera
        else:
            raise ValueError(f"Chimera node labeling {labels_s} not recognized")

        make_mapping = _chimera_pegasus_sublattice_mapping

    elif source.graph.get('family') == 'pegasus':
        m_s = source.graph['rows']
        if offset_list is None:
            ranges = range(m_t - m_s + 1), range(m_t - m_s), range(m_t - m_s)
            offset_list = ((t, y, x) for t in range(3) for y in ranges[t] for x in ranges[t])

        labels_s = source.graph['labels']
        if labels_s == 'int':
            source_to_inner = _pegasus_coordinates_cache[m_s].linear_to_nice
        elif labels_s == 'coordinate':
            source_to_inner = _pegasus_coordinates_cache[m_s].pegasus_to_nice
        elif labels_s == 'nice':
            def source_to_inner(q):
                return q
        else:
            raise ValueError(f"Pegasus node labeling {labels_s} not recognized")

        make_mapping = _pegasus_pegasus_sublattice_mapping

    else:
        raise ValueError("source graph must be a Chimera graph or Pegasus graph constructed by dwave_networkx.chimera_graph or dwave_networkx.pegasus_graph respectively")

    for offset in offset_list:
        yield make_mapping(source_to_inner, nice_to_target, offset)




# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import unittest

import networkx as nx
import dwave_networkx as dnx

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = False

try:
    import numpy as np
except ImportError:
    np = False

_display = os.environ.get('DISPLAY', '') != ''


@unittest.skipUnless(np and plt, "No numpy or matplotlib")
class TestDrawing(unittest.TestCase):
    def test_pegasus_layout_coords(self):
        G = dnx.pegasus_graph(2, coordinates=True)
        pos = dnx.pegasus_layout(G)

    def test_pegasus_layout_ints(self):
        G = dnx.pegasus_graph(2)
        pos = dnx.pegasus_layout(G)

    def test_pegasus_layout_chim(self):
        G = dnx.pegasus_graph(2, nice_coordinates=True)
        pos = dnx.pegasus_layout(G)

    def test_pegasus_layout_ints_nodata(self):
        G = dnx.pegasus_graph(2, data=False)
        pos = dnx.pegasus_layout(G)

    def test_pegasus_layout_crosses(self):
        G = dnx.pegasus_graph(2)
        pos = dnx.pegasus_layout(G, crosses=True)

    def test_pegasus_layout_ints_badcenter(self):
        G = dnx.pegasus_graph(2, data=False)
        with self.assertRaises(ValueError):
            pos = dnx.pegasus_layout(G, center=(0, 0, 0, 0))

    def test_pegasus_layout_ints_noinfo(self):
        G = dnx.pegasus_graph(2, data=False)
        badG = nx.Graph()
        badG.add_edges_from(G.edges())
        with self.assertRaises(ValueError):
            pos = dnx.pegasus_layout(badG)

    @unittest.skipUnless(_display, " No display found")
    def test_draw_pegasus_yield(self):
        G = dnx.pegasus_graph(3, data=False)
        G.remove_edges_from([(5,104),(12,96),(23,112)])
        G.remove_nodes_from([109,139])
        dnx.draw_pegasus_yield(G)

    @unittest.skipUnless(_display, " No display found")
    def test_draw_pegasus_biases(self):
        G = dnx.pegasus_graph(2)
        h = {v: v % 12 for v in G}
        J = {(u, v) if u % 2 else (v, u): (u+v) % 24 for u, v in G.edges()}
        for v in G:
            J[v, v] = .1

        dnx.draw_pegasus(G, linear_biases=h, quadratic_biases=J)

    @unittest.skipUnless(_display, " No display found")
    def test_draw_pegasus_embedding(self):
        P = dnx.pegasus_graph(2)
        G = nx.grid_graph([3, 3, 2])
        emb = {(0, 0, 0): [35], (0, 0, 1): [12], (0, 0, 2): [31], (0, 1, 0): [16],
               (0, 1, 1): [36], (0, 1, 2): [11], (0, 2, 0): [39], (0, 2, 1): [6],
               (0, 2, 2): [41], (1, 0, 0): [34], (1, 0, 1): [13], (1, 0, 2): [30],
               (1, 1, 0): [17], (1, 1, 1): [37], (1, 1, 2): [10], (1, 2, 0): [38],
               (1, 2, 1): [7], (1, 2, 2): [40]}
        dnx.draw_pegasus_embedding(P, emb)
        dnx.draw_pegasus_embedding(P, emb, embedded_graph=G)
        dnx.draw_pegasus_embedding(P, emb, interaction_edges=P.edges())
        dnx.draw_pegasus_embedding(P, emb, crosses=True)

    @unittest.skipUnless(_display, " No display found")
    def test_draw_overlapped_chimera_embedding(self):
        C = dnx.pegasus_graph(2)
        emb = {0: [12, 35], 1: [12, 31], 2: [32], 3: [14]}
        dnx.draw_pegasus_embedding(C, emb, overlapped_embedding=True)
        dnx.draw_pegasus_embedding(C, emb, overlapped_embedding=True, show_labels=True)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dwave_networkx.algorithms.independent_set import *
from dwave_networkx.algorithms.canonicalization import *
from dwave_networkx.algorithms.clique import *
from dwave_networkx.algorithms.cover import *
from dwave_networkx.algorithms.matching import *
from dwave_networkx.algorithms.social import *
from dwave_networkx.algorithms.elimination_ordering import *
from dwave_networkx.algorithms.coloring import *
from dwave_networkx.algorithms.max_cut import *
from dwave_networkx.algorithms.markov import *
from dwave_networkx.algorithms.tsp import *
from dwave_networkx.algorithms.partition import *



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ["maximum_weighted_independent_set",
           "maximum_weighted_independent_set_qubo",
           "maximum_independent_set",
           "is_independent_set",
           ]


@binary_quadratic_model_sampler(2)
def maximum_weighted_independent_set(G, weight=None, sampler=None, lagrange=2.0, **sampler_args):
    """Returns an approximate maximum weighted independent set.

    Defines a QUBO with ground states corresponding to a
    maximum weighted independent set and uses the sampler to sample
    from it.

    An independent set is a set of nodes such that the subgraph
    of G induced by these nodes contains no edges. A maximum
    independent set is an independent set of maximum total node weight.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximum cut weighted independent set.

    weight : string, optional (default None)
        If None, every node has equal weight. If a string, use this node
        attribute as the node weight. A node without this attribute is
        assumed to have max weight.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.
        
    lagrange : optional (default 2)
        Lagrange parameter to weight constraints (no edges within set) 
        versus objective (largest set possible).

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    indep_nodes : list
       List of nodes that form a maximum weighted independent set, as
       determined by the given sampler.

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Independent Set on Wikipedia <https://en.wikipedia.org/wiki/Independent_set_(graph_theory)>`_

    `QUBO on Wikipedia <https://en.wikipedia.org/wiki/Quadratic_unconstrained_binary_optimization>`_

    .. [AL] Lucas, A. (2014). Ising formulations of many NP problems.
       Frontiers in Physics, Volume 2, Article 5.

    """
    # Get a QUBO representation of the problem
    Q = maximum_weighted_independent_set_qubo(G, weight, lagrange)

    # use the sampler to find low energy states
    response = sampler.sample_qubo(Q, **sampler_args)

    # we want the lowest energy sample
    sample = next(iter(response))

    # nodes that are spin up or true are exactly the ones in S.
    return [node for node in sample if sample[node] > 0]


@binary_quadratic_model_sampler(1)
def maximum_independent_set(G, sampler=None, lagrange=2.0, **sampler_args):
    """Returns an approximate maximum independent set.

    Defines a QUBO with ground states corresponding to a
    maximum independent set and uses the sampler to sample from
    it.

    An independent set is a set of nodes such that the subgraph
    of G induced by these nodes contains no edges. A maximum
    independent set is an independent set of largest possible size.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a maximum cut independent set.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.
        
    lagrange : optional (default 2)
        Lagrange parameter to weight constraints (no edges within set) 
        versus objective (largest set possible).

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    indep_nodes : list
       List of nodes that form a maximum independent set, as
       determined by the given sampler.

    Example
    -------
    This example uses a sampler from
    `dimod <https://github.com/dwavesystems/dimod>`_ to find a maximum
    independent set for a graph of a Chimera unit cell created using the
    `chimera_graph()` function.

    >>> import dimod
    >>> sampler = dimod.SimulatedAnnealingSampler()
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> indep_nodes = dnx.maximum_independent_set(G, sampler)

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    References
    ----------

    `Independent Set on Wikipedia <https://en.wikipedia.org/wiki/Independent_set_(graph_theory)>`_

    `QUBO on Wikipedia <https://en.wikipedia.org/wiki/Quadratic_unconstrained_binary_optimization>`_

    .. [AL] Lucas, A. (2014). Ising formulations of many NP problems.
       Frontiers in Physics, Volume 2, Article 5.

    """
    return maximum_weighted_independent_set(G, None, sampler, lagrange, **sampler_args)


def is_independent_set(G, indep_nodes):
    """Determines whether the given nodes form an independent set.

    An independent set is a set of nodes such that the subgraph
    of G induced by these nodes contains no edges.

    Parameters
    ----------
    G : NetworkX graph
       The graph on which to check the independent set.

    indep_nodes : list
       List of nodes that form a maximum independent set, as
       determined by the given sampler.

    Returns
    -------
    is_independent : bool
        True if indep_nodes form an independent set.

    Example
    -------
    This example checks two sets of nodes, both derived from a
    single Chimera unit cell, for an independent set. The first set is
    the horizontal tile's nodes; the second has nodes from the horizontal and
    verical tiles.

    >>> import dwave_networkx as dnx
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> dnx.is_independent_set(G, [0, 1, 2, 3])
    True
    >>> dnx.is_independent_set(G, [0, 4])
    False

    """
    return len(G.subgraph(indep_nodes).edges) == 0


def maximum_weighted_independent_set_qubo(G, weight=None, lagrange=2.0):
    """Return the QUBO with ground states corresponding to a maximum weighted independent set.

    Parameters
    ----------
    G : NetworkX graph

    weight : string, optional (default None)
        If None, every node has equal weight. If a string, use this node
        attribute as the node weight. A node without this attribute is
        assumed to have max weight.
        
    lagrange : optional (default 2)
        Lagrange parameter to weight constraints (no edges within set) 
        versus objective (largest set possible).

    Returns
    -------
    QUBO : dict
       The QUBO with ground states corresponding to a maximum weighted independent set.

    Examples
    --------

    >>> from dwave_networkx.algorithms.independent_set import maximum_weighted_independent_set_qubo
    ...
    >>> G = nx.path_graph(3)
    >>> Q = maximum_weighted_independent_set_qubo(G, weight='weight', lagrange=2.0)
    >>> Q[(0, 0)]
    -1.0
    >>> Q[(1, 1)]
    -1.0
    >>> Q[(0, 1)]
    2.0

    """

    # empty QUBO for an empty graph
    if not G:
        return {}

    # We assume that the sampler can handle an unstructured QUBO problem, so let's set one up.
    # Let us define the largest independent set to be S.
    # For each node n in the graph, we assign a boolean variable v_n, where v_n = 1 when n
    # is in S and v_n = 0 otherwise.
    # We call the matrix defining our QUBO problem Q.
    # On the diagnonal, we assign the linear bias for each node to be the negative of its weight.
    # This means that each node is biased towards being in S. Weights are scaled to a maximum of 1.
    # Negative weights are considered 0.
    # On the off diagnonal, we assign the off-diagonal terms of Q to be 2. Thus, if both
    # nodes are in S, the overall energy is increased by 2.
    cost = dict(G.nodes(data=weight, default=1))
    scale = max(cost.values())
    Q = {(node, node): min(-cost[node] / scale, 0.0) for node in G}
    Q.update({edge: lagrange for edge in G.edges})

    return Q
